# -*- coding: utf-8 -*-
"""
Created on Sun May 21 08:51:33 2023

@author: atakan
"""


import copy

# import src.models.fluids.fluid_props as fprop
import numpy as np
from scipy.optimize import minimize  # root, root_scalar
import matplotlib.pyplot as plt
import carbatpy as cb


class StaticHeatExchanger:
    """
    Class for static counter-flow heat exchanger

    means: no time dependence and no heat transfer coefficients * areas
    are used (UA)! Instead a minimum approach temperature is tried to be met.
    At the moment, this is mainly done by varying one of the mass flow rates.
    But this is sometimes not enough and a variation of the working fluid
    pressure will soon be included.
    Only the first law and second law will be checked (the latter must be
    improved).

    """

    def __init__(self, fluids, h_dot_min, h_out_w, h_limit_s=np.nan,
                 points=50, d_temp_separation_min=0.5, calc_type="const",
                 name="evaporator"):
        """
        class to calculate (static/steady state) heat-exchangers

        includes pinch-point analysis and plotting,
        only implemented for simple thermodynamic calculations
        (no convection coefficients and heat exchanger areas regarded yet)

        Parameters
        ----------
        fluids : list of 2 fprop.Fluid
            The definition of the two fluids, as they enter the heat exchangers.
            Typically at room temperature.
        h_dot_min : float
            enthalpy flow rate (W) which has to be transfered.
        h_out_w : float
            exit enthalpy of the working fluid.
        h_limit_s : float or nan, optional
            if there is a limit in enthalpy for the secondary fluid, it can be
            given here. The default is nan.
        points : integer, optional
            for how many points (array) shall the minimum approach temperature
            be checked and properties be returned (for plotting etc.)
            The default is 30.
        d_temp_separation_min : float, optional
            Minimium approach temperature (pinch point) between the two fluids.
            The default is 0.5.
        calc_type : string, optional
            which calculation type shall be performed; only one implemented so
            far. The default is "const".
        name : string, optional
            name of the heat exchanger. The default is "evaporator".

        Returns
        -------
        None.

        """
        self.fluids = fluids
        state_in_w = fluids[0].properties.state
        self.m_dot_s = 0
        self.h_limit_s = h_limit_s
        self.q_dot = h_dot_min
        self.m_dot_w = np.abs(h_dot_min/(h_out_w-state_in_w[2]))
        self.h_out_w = h_out_w
        self.h_out_s = float(copy.copy(h_limit_s))  # this may be varied
        self.points = points
        self. d_temp_separation_min = d_temp_separation_min
        self.heating = -1
        if h_out_w < state_in_w[2]:
            self.heating = 1  # condenser (heating of the secondary fluid)
        self.calc_type = calc_type
        self.name = name
        self.all_states = np.zeros(
            (points, len(cb.fprop._THERMO_STRING.split(";"))))
        self.h_in_out = np.zeros((2, 4))
        self.dt_mean = None
        self.dt_min = None
        self.dt_max = None
        self.warning = 0
        self.warning_message = "All o.k."

    def pinch_calc(self,  verbose=False, plot_info={}):
        """
        Calculate the changes in enthalpy and temperature in the heat exchanger

        counter-flow hex assumed! Both flows are isobaric.
        Is used to check, whether the second law is violated. The factor can
        be used to vary the mass flow rate of the working fluid, until no
        violation is found (done in root finding).

        Parameters
        ----------
        verbose : boolean, optional
            if True, several variables will be printed out. Default is False.
        plot_info : dictionary, optional
            if not empty a Figure, an Axes, a list of What shall be plotted,
            a list with the colour/styles and a list with the labels must be 
            passed. in "what", the two numbers coincide with the fluid THERMO
            order. The x-shift can be used in cycle calculations, to shift the
            curves, by the value (it will be added).
            The names in the dictionary are: "fig", "ax", "what","col",
            "label", "x-shift". Default is empty.


        Raises
        ------
        Exception
            if temperatures are not consistent.

        Returns
        -------
        m_dot_s : float
            secondary fluid mass flow rate (kg/s.
        d_tempall : numpy-array
            The temperature differences along the counter-flow heat exchanger.
        w_array : array
            properties of the working fluid along the heat exchanger
            (T,p,h, etc. see fluid class).
        s_array : array
            properties of the secondary fluid along the heat exchanger
            (T,p,h, etc. see fluid class).

        """
        self.warning = 0
        self.warning_message = "All o.k."
        w_in = copy.copy(self.fluids[0])
        s_in = copy.copy(self.fluids[1])

        w_out = copy.copy(self.fluids[0])  # not yet the correct state!
        s_out = copy.copy(self.fluids[1])

        #  fixed values
        self. h_in_out[1, 0] = s_in.properties.enthalpy
        self. h_in_out[0, 0] = w_in.properties.enthalpy
        self. h_in_out[0, 1] = self.h_out_w

        # fixed limiting state, secondary fluid
        state_out_s = s_out.set_state([self.h_out_s,
                                       s_in.properties.pressure], "HP")

        self. h_in_out[1, 1] = state_out_s[2]

        h_delta_s = np.abs(
            state_out_s[2] - s_in.properties.enthalpy)

        # fixed heat flow,  determines mass flow rate
        self.m_dot_s = self.q_dot / h_delta_s

        h_array = np.linspace(
            self.h_in_out[1, 0], self.h_in_out[1, 1], self.points)
        values = np.zeros((self.points, 2))
        values[:, 0] = h_array
        values[:, 1] = s_out.properties.pressure

        s_array = s_out.set_state_v(values, given="HP")
        # if not(self.heating): s_array =np.flip(s_array, axis =0)

        # now working fluid:

        h_array = np.linspace(self.h_in_out[0, 1],
                              self.h_in_out[0, 0],
                              self.points)

        values = np.zeros((self.points, 2))
        values[:, 0] = h_array.T
        values[:, 1] = w_out.properties.pressure

        w_array = w_out.set_state_v(values, "HP")
        # temperature difference, ok?
        d_tempall = w_array[:, 0]-s_array[:, 0]
        self.dt_mean = d_tempall.mean()
        self.dt_min = np.abs(d_tempall).min()
        self.dt_max = np.abs(d_tempall).max()

        if abs(self.dt_min) - self.d_temp_separation_min < -1e-3:
            self.warning = 900
            self.warning_message = "Below minumum approch temperature!"
        positive = np.any(d_tempall > 0)
        negative = np.any(d_tempall < 0)
        below = True
        if self.heating < 0:
            below = False

        crossing = (positive > 0 and negative > 0)
        wrong_side = (positive > 0 and not below) or (negative > 0 and below)
        if crossing or wrong_side:
            self.warning = 999
            self.dt_mean = 1e6
            self.warning_message = "Temperatures crossing or wrong side!"

        # if self.heating:
        #     d_tempall = w_array[:, 0]-s_array[:, 0]
        # else:
        #     d_tempall = w_array[:, 0] - np.flip(s_array[:, 0],axis=0)
        
        if len(plot_info)>0:
            if plot_info["what"][0] ==2:
                data = (w_array[:,plot_info["what"][0]] -
                        w_array[:,plot_info["what"][0]].min()) \
                    *self.m_dot_w + plot_info["x-shift"][0]
                plot_info["ax"].plot(data,
                                     w_array[:,plot_info["what"][1]],
                                     plot_info["col"][0],
                                     label=plot_info["label"][0])
                data = (s_array[:,plot_info["what"][0]] -
                        s_array[:,plot_info["what"][0]].min()) \
                    *self.m_dot_s + plot_info["x-shift"][1]
                plot_info["ax"].plot(data,
                                     s_array[:,plot_info["what"][1]],
                                     plot_info["col"][1],
                                     label=plot_info["label"][1])
            else:
                print(f"H-Ex: plotting only implemented fot T-H_dot [2,0]. You requested{plot_info['what']}")
        if verbose:
            print(
                f"Min T-distance {self.dt_min}, Mean T-distance {self.dt_mean}")
            if self.heating > 0:
                print("cond", d_tempall[0], d_tempall[-1],
                      d_tempall.min(), d_tempall.max())
            else:
                print("evap", d_tempall[0], d_tempall[-1],
                      d_tempall.max(), d_tempall.min())

        return self.m_dot_s, d_tempall*self.heating, w_array, s_array

    def pinch_root(self, h_out_s,  secondary, verbose=False):
        """
        Function for finding the minimum mean temperature difference in a heat
        exchanger, while not going below the minimum approach temperature.

        The output enthalpy of the secondary fluid is set (as default) and the
        heat exchanger is evaluated.


        Parameters
        ----------
        h_out_s : float
            the output enthalpy, default would be of the secondary fluid.
        secondary : boolean
            shall the secondary fluid output be varied (or the working fluid)?

        Returns
        -------
        mean T-difference, float
            root tries to reach a value of 0.

        """
        
        if isinstance(h_out_s, float): # numpy expects a float (later)
            value = h_out_s
        else:
           value = h_out_s[0]
           
        if secondary:
            self.h_out_s = value
        else:
            self.h_out_w = value
            
        mdot_s, d_temps, wf_states, sf_states = self.pinch_calc()

        if self.warning < 100:
            return abs(self.dt_mean)
        else:
            return 500.0

    def find_pinch(self, secondary=True):
        """
        Function tries to vary the secondary fluid enthalpy  until a
        minimum approach temperature is reached. This also changes the
        mass flow rate.  This is then also the new
        exit state
        within the heat exchanger. If this is also not succesful,
        self.warning is set to 1. This should be checked.
        Parameters
        ----------
        secondary : Boolean, optional
            shall the output state of the secondary (True) or the working
            (False) fluid be varied? default = True

        Returns
        -------
        float
            the optimized enthalpy of the secondary fluid.

        """
        verbose = False
        if secondary:
            x0 = copy.copy(self.h_out_s)
        else:
            x0 = copy.copy(self.h_out_w)
        
        tolerance = 1e-3

        try:
            result = minimize(self.pinch_root, x0, args =(secondary,),
                              method='Nelder-Mead', tol=tolerance)

            if verbose:
                print(
                    f"result {result}, heating {self.heating}")

            if result.success or result.status == 2:
                if result.status == 2:
                    self.warning = 2  # T-difference probably smaller
                    self.warning_message = "Minimization problem: "+result.message
                if secondary:
                    self.h_out_s = result.x[0]
                else:
                    self.h_out_w = result.x[0]
                print(
                    f"Min T-distance {self.dt_min:.3f}, Mean T-distance {self.dt_mean:.3f}")

                return result.x[0]

        # except:
        except Exception as inst:
            print("find pinch:", type(inst))    # the exception type
            print(inst.args)     # arguments stored in .args
            print(inst)          # __str__ allows args to be printed directly,
            # but may be overridden in exception subclasses

            
            print("root-exception", self.heating)
            return 10000
        print("root-finding problem! (in heat_exchanger_thermo_v2.find_pinch)",
              result)
        print(f"Heating: {self.heating}")
        self.warning = 1
        return self.warning

    def pinch_plot(self, plot_fname="", plotting=True):
        """
        calculates the secondary fluid output state and mass flow, for the
        minimum approach temperature of the HeatExchanger instance. When wanted,
        this is also plotted

        Parameters
        ----------
        plot_fname : string, optional
            file-name to store the plot. The default is "".
        plotting : Boolean, optional
            should it be plotted? The default is True.

        Returns
        -------
        m_dot_s : float
            mass flow raete of the secondary fluid in SI units (kg/s).
        d_tempall : np.array
            the temperature differences between the two fluids along the heat
            exchanger.
        w_array : np.array [self.points, 7]
            the states of the working fluid along the heat exchanger.
        s_array : np.array [self.points, 7]
            the states of the secondary fluid along the heat exchanger.

        """
        print(f"------pinch-plot running -----plot:{plotting}")
        m_dot_s, d_tempall, w_array, s_array = self.pinch_calc()

        if plotting:
            h_w_plot_array = (
                w_array[:, 2] - w_array[:, 2].min()) * self.m_dot_w
            fig, ax_one = plt.subplots(1, 1)
            ax_one.plot((s_array[:, 2] - s_array[:, 2].min()) * self.m_dot_s,
                        s_array[:, 0], "v")
            ax_one.plot(h_w_plot_array, w_array[:, 0], "o")
            ax_one.set_xlabel(
                "specific enthalpy flow per mass of secondary fluid / (J / kg)")
            ax_one.set_ylabel("temperature / (K)")
            ax_one.set_title("heat exchanger, simple")
        if plot_fname != "":
            fig.savefig(plot_fname)

        return m_dot_s, d_tempall, w_array, s_array


if __name__ == "__main__":
    # two test cases condenser and evaporator:

    FLUID = "Propane * Pentane"  # working fluid
    FLS = "Methanol"  # "Water"  # secondary fluid
    comp = [.50, 0.5]
    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])
    D_TEMP_MIN = 6.0

    # Condenser, working fluid fixes all, secondary output enthalpy can be varied:
    SEC_TEMP_IN = 300.0
    SEC_TEMP_OUT_MAX = 370.0
    SEC_PRES_IN = 5e5
    H_DOT = 1e3
    state_sec_out = secFluid.set_state([SEC_TEMP_OUT_MAX, SEC_PRES_IN], "TP")

    state_sec_in = secFluid.set_state(
        [SEC_TEMP_IN, SEC_PRES_IN], "TP")  # this is the entering state

    # working fluid

    TEMP_SAT_VAP = SEC_TEMP_OUT_MAX + D_TEMP_MIN
    state_in = myFluid.set_state(
        [TEMP_SAT_VAP, 1.], "TQ")  # find minimum pressure

    WF_TEMP_IN = TEMP_SAT_VAP + D_TEMP_MIN
    WF_TEMP_OUT = SEC_TEMP_IN + D_TEMP_MIN
    state_out = myFluid.set_state([WF_TEMP_OUT, state_in[1]], "TP")

    state_in = myFluid.set_state([myFluid.properties.pressure,
                                  WF_TEMP_IN],
                                 "PT")

    hex0 = StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
                               state_sec_out[2],
                               d_temp_separation_min=D_TEMP_MIN)
   
    factor0 = hex0.find_pinch()
    if hex0.warning > 0:
        print(hex0.warning_message)
        
    # now plotting can directly be done in pinch_calc 2024-05-24
    fig_act, ax_act = plt.subplots(1)
    PLOT_INFO ={"fig":fig_act, "ax":ax_act, "what":[2,0],"col":["r:","ko"],
    "label":["work,c","sec,c"], "x-shift":[0,0]}
    hex0.pinch_calc(plot_info=PLOT_INFO)
    ax_act.legend()
    #-----------------------------------------------
    #       Previous way
    # ms0, d_tempall0, w0, s0 = hex0.pinch_plot("hex-plot.png")
    

    #  Evaporator: ----------------------------

    SEC_TEMP_IN = 300.0
    SEC_TEMP_OUT = 285
    SEC_PRES_IN = 15e5
    H_DOT = 1e3
    extra = 2
    # D_TEMP_SUPER = 5.
    D_TEMP_MIN = 6.0
    state_sec_out = secFluid.set_state([SEC_TEMP_OUT, SEC_PRES_IN], "TP")
    # this mus be the last set_state before the hex is constructed:
    state_sec_in = secFluid.set_state([SEC_TEMP_IN, SEC_PRES_IN], "TP")

    # WF_TEMP_IN = SEC_TEMP_OUT  # - D_TEMP_MIN
    state_out = myFluid.set_state([SEC_TEMP_IN-D_TEMP_MIN - extra, 1.0], "TQ")
    state_in = myFluid.set_state(
        [SEC_TEMP_OUT-D_TEMP_MIN - extra, state_out[1]], "TP")

    # print("state in", state_in)

    hex1 = StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
                               state_sec_out[2],
                               d_temp_separation_min=D_TEMP_MIN)
    # ms1, d_tempall1, w1, s1 = hex1.pinch_calc()
    

    factor_out = hex1.find_pinch()
    if hex1.warning > 2:
        print("Second heat exchanger:", hex1.warning_message)
    else:

        # plotting in the same figure
        PLOT_INFO ={"fig":fig_act, "ax":ax_act, "what":[2,0],"col":["k:","bo"],
        "label":["work,e","sec,e"], "x-shift":[0,0]}
        hex1.pinch_calc(plot_info=PLOT_INFO)
        ax_act.legend()



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 16 11:13:12 2023

@author: atakan
"""

import numpy as np
import carbatpy as cb

import pandas as pd
import seaborn as sbn
import matplotlib.pyplot as plt
import datetime
dir_name = cb._RESULTS_DIR + "\\optimal_hp_fluid"
fname = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M")+"threeAlkanes"
plt.style.use('seaborn-v0_8-poster')  # 'seaborn-v0_8-poster')

FLUID_COMP = "Propane * Butane * Pentane"  # " * Hexane"
comp = [.5, 0., 0.5]  # , 0.0]
# comp =[0.000,1.0,0.00,0.0]

FLS = "Water"  #
FLCOLD = "Methanol"  # "Water"  #

flm = cb.fprop.fluid_model(FLUID_COMP)
my_fluid = cb.fprop.Fluid(flm, comp)

secFlm = cb.fprop.fluid_model(FLS)
secondary_fluid = cb.fprop.Fluid(secFlm, [1.])

coldFlm = cb.fprop.fluid_model(FLCOLD)
cold_fluid_act = cb.fprop.Fluid(coldFlm, [1.])

# Condenser(c) and storage (s), secondary fluids fix all, temperatures(T in K),
# pressures (p in Pa)

SECONDARY_TEMP_IN = 290.0
COLD_TEMP_IN = SECONDARY_TEMP_IN
SECONDARY_TEMP_OUT = 363.  # 395.0
COLD_TEMP_OUT = 257.
SP_IN = 5E5
CP_IN = 5E5
DH_MIN = 1E3  # AND HEAT_FLOW RATE (W)
ETA_S = 0.65
D_TEMPMIN = 1.530  # MINIMUM APPROACH TEMPERATURE (PINCH POINT)
Q_DOT_H = 3000.

state_sec_in = secondary_fluid.set_state([SECONDARY_TEMP_IN, SP_IN], "TP")
state_cold_in = cold_fluid_act.set_state([COLD_TEMP_IN, CP_IN], "TP")

# working fluid
D_TEMPSUPER = 5.  # super heating of working fluid
TEMP_DEW = SECONDARY_TEMP_OUT
state_in_act = my_fluid.set_state([TEMP_DEW, 1.], "TQ")  # find minimum pressure

TEMP_IN = TEMP_DEW + D_TEMPSUPER
state_in_act = my_fluid.set_state([my_fluid.properties.pressure,
                              my_fluid.properties.temperature + D_TEMPSUPER], "PT")
p_high_act = state_in_act[1]

state_in_act = my_fluid.set_state(
    [SECONDARY_TEMP_IN - D_TEMPSUPER, 1.], "TQ")  # find minimum pressure
p_low_act = state_in_act[1]

state_in_act = my_fluid.set_state([p_low_act,
                              SECONDARY_TEMP_IN - D_TEMPMIN], "PT")
print(f"p-ratio: {p_high_act/p_low_act: .2f}, p_low: {p_low_act/1e5: .2} bar")

state_in_act = my_fluid.set_state([p_low_act, TEMP_IN], "PT")

fixed_points_act = {"eta_s": ETA_S,
                "p_low": p_low_act,
                "temp_hh": SECONDARY_TEMP_OUT,
                "temp_hl": SECONDARY_TEMP_IN,
                "temp_lh": SECONDARY_TEMP_IN,
                "temp_ll": COLD_TEMP_OUT,
                "q_dot_h": Q_DOT_H,
                "d_tempmin": D_TEMPMIN}

hp0_act = cb.hp_simple.heat_pump([my_fluid, secondary_fluid, cold_fluid_act],
                             fixed_points_act)
cop_act = hp0_act.calc_p_high(p_high_act)
hp0_act.hp_plot()
out = hp0_act.evaluation
print(
    f"COP: {out[0]/out[1]:.2f},p-ratio: { out[2]/out[3]:.2f}, p_low {out[3]/1e5:.2f} bar")


def hp_comp_cop(composition, work_fluid, sec_fluid, cold_fluid,
                fixed_points, d_tempsuper=5.0, opt=+1):
    # working fluid
    sum3 = np.sum(composition)
    if sum3 > 1:
        return 0

    last = 1-sum3
    new_composition = np.array([*composition, last])

    work_fluid.set_composition(new_composition)
    temp_dew = fixed_points["temp_hh"]
    state_in = work_fluid.set_state(
        [temp_dew, 1.], "TQ")  # find minimum pressure

    # TEMP_IN = temp_dew + D_TEMPSUPER
    state_in = work_fluid.set_state([work_fluid.properties.pressure,
                                    work_fluid.properties.temperature + D_TEMPSUPER], "PT")
    p_high = state_in[1]

    state_in = work_fluid.set_state(
        [fixed_points["temp_lh"] - d_tempsuper, 1.], "TQ")  # find minimum pressure
    p_low = state_in[1]

    state_in = my_fluid.set_state([p_low,
                                  SECONDARY_TEMP_IN - D_TEMPMIN], "PT")
    fixed_points["p_low"] = p_low
    hp0 = cb.hp_simple.heat_pump(
        [work_fluid, sec_fluid, cold_fluid], fixed_points)
    cop = hp0.calc_p_high(p_high)
    print("Result-hp-opt:", new_composition, np.sum(new_composition),
          cop, p_low, p_high, p_high/p_low)
    if opt == -1:
        return cop * opt

    return cop, p_low/1e5, p_high/1e5


all_res = []
x0_range = np.linspace(0.0, 0.6, 12)
x1_range = np.linspace(0, 0.6, 10)
for x0 in x0_range:
    for x1 in x1_range:
        if x0+x1 <= 1:
            x2 = 1-(x0+x1)
            comp = np.array([x0, x1])
            results = hp_comp_cop(comp, my_fluid, secondary_fluid,
                                  cold_fluid_act,
                                  fixed_points_act,
                                  d_tempsuper=D_TEMPSUPER,
                                  opt=+1)
            if results[0] > 1:
                new = np.array([*results, x0, x1, x2,
                                *fixed_points_act.values()])
                all_res.append(new)

names = ["COP", "$p_{low}$", "$p_{high}$", "$x_{Propane}$",
         "$x_{Butane}$", "$x_{Pentane}$", *fixed_points_act.keys()]
all_res = np.array(all_res)
dframe = pd.DataFrame(all_res, columns=names)
dframe.to_csv(dir_name+"\\"+fname+".csv")
dframe2 = dframe.round(3)
f, ax = plt.subplots(figsize=(10, 6), layout="constrained")
fff = sbn.scatterplot(x=names[4], y=names[5],
                      hue=names[2], size=names[0], data=dframe2, ax=ax)
sbn.move_legend(fff, "upper left", bbox_to_anchor=(1, 1))
f.savefig(dir_name+"\\"+fname+".png")
# bounds = ((0.05, 0.95), (0.00, 0.65))


# def constr(comp):
#     print(comp, np.sum(comp))
#     return 1-np.sum(comp)

# res = opt.differential_evolution(hp_comp_cop, bounds,
#                args=(my_fluid, secFluid, cold_fluid_act, fixed_points)
#                )
# print(res)



# -*- coding: utf-8 -*-
"""
Helpers for file handling, e.g. copy the actual py-File after
adding the actual date-time and copy it to a directory


Created on Thu Feb  1 16:18:03 2024

@author: atakan
"""
import os
import shutil
import datetime

print(__file__)


def copy_script_add_date(file_name_end, source_file, directory):
    """
    File handling, copy the actual py-File after
    adding the actual date-time and copy it to a directory which is also named
    date-time - file_name_end
    return the date-time for further usage

    Parameters
    ----------
    file_name_end : string
        the file name to be added to the date, also for the directory name.
    source_file : string
        The source file to copy (from __file__).
    directory : string
        The directory to copy to.

    Returns
    -------
    directory : String
        The new directory name and the file name(s) without extension.

    """
    py_name = source_file.rsplit("\\", maxsplit=1)[-1]
    access = 0o777
    current_date = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M")

    output_folder = f"{current_date}-{file_name_end}"
    
    directory = os.path.join(directory,output_folder)
    os.makedirs(directory, access)
    os.chdir(directory)
    
    destination_file = os.path.join(directory, current_date + py_name)
    shutil.copy(__file__, destination_file)
    return os.path.join(directory,output_folder)


if __name__ == "__main__":
    
    from carbatpy import _RESULTS_DIR
    END_NAME = "test"
    print(copy_script_add_date(END_NAME, __file__, _RESULTS_DIR))



# -*- coding: utf-8 -*-
"""
Created on Tue May 21 13:56:39 2024

@author: atakan
"""


import copy
import numpy as np
import matplotlib.pyplot as plt

import carbatpy as cb

_RESULTS_ = cb._RESULTS_DIR

cb._T_SURROUNDING =20+273.15
if __name__ == "__main__":
    # -------HEAT PUMP------------------------------

    FLUID = "Propane * Butane * Pentane * Ethane"
    # comp = [.15, 0.55, 0.0, 0.30]
    # comp = [0.4,	0.3,	0.3, 0.0]  # [0.164,.3330,.50300,0.0]
    # comp = [.05, 0.65, 0.0, 0.30]
    comp = [0.7,	0.1,	0.2, 0.0] 
    FLS = "Water"  #
    FLCOLD = "Methanol"  # "Water"  #

    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])

    coldFlm = cb.fprop.FluidModel(FLCOLD)
    coldFluid = cb.fprop.Fluid(coldFlm, [1.])

    # Condenser(c) and storage (s), secondary fluids fix all, temperatures(T in K),
    # pressures (p in Pa)
    _ETA_S_ = 0.6  # interesting when changed from 0.69 to 0.65, the efficiency
    # decreases, the reason is the low quality along throtteling then
    T_MAX = 77 + 273.15 # 369.
    T_MIN = 253.15
    _STORAGE_T_IN_ = cb._T_SURROUNDING
    _COLD_STORAGE_T_IN_ = _STORAGE_T_IN_
    _STORAGE_T_OUT_ = T_MAX  # 395.0
    _COLD_STORAGE_T_OUT_ = T_MIN
    _STORAGE_P_IN_ = 5e5  # pressure
    _COLD_STORAGE_P_IN_ = 5e5
    _Q_DOT_MIN_ = 1e3  # heat_flow rate to storage (W)
    _D_T_SUPER_ = 3.8  # super heating of working fluid
    _D_T_MIN_ = 2.  # minimum approach temperature (pinch point)
    _D_T_EVAP = 5.0
    _D_T_COND = -7
    # high T-storages
    state_sec_out = secFluid.set_state([_STORAGE_T_OUT_, _STORAGE_P_IN_], "TP")
    state_sec_in = secFluid.set_state([_STORAGE_T_IN_, _STORAGE_P_IN_], "TP")

    #  low T storages:
    state_cold_out = coldFluid.set_state(
        [_COLD_STORAGE_T_OUT_, _COLD_STORAGE_P_IN_], "TP")
    state_cold_in = coldFluid.set_state(
        [_COLD_STORAGE_T_IN_, _COLD_STORAGE_P_IN_], "TP")

    # working fluid
    T_DEW = _STORAGE_T_OUT_  +_D_T_COND# + _D_T_MIN_
    state_in_cond = myFluid.set_state([T_DEW, 1.], "TQ")  # find high pressure
    state_out_cond = myFluid.set_state([_STORAGE_T_IN_ + _D_T_MIN_,
                                        state_in_cond[1]], "TP")
    state_satv_evap = myFluid.set_state(
        [_STORAGE_T_IN_-_D_T_MIN_-_D_T_SUPER_-_D_T_EVAP, 1.], "TQ")  # find minimum pressure
    p_low = state_satv_evap[1]

    T_IN = _STORAGE_T_IN_ - _D_T_MIN_

    state_out_evap = myFluid.set_state([p_low,
                                        T_IN], "PT")

    FIXED_POINTS = {"eta_s": _ETA_S_,
                    "p_low": state_out_evap[1],
                    "p_high": state_in_cond[1],
                    "T_hh": _STORAGE_T_OUT_,
                    "h_h_out_sec": state_sec_out[2],
                    "h_h_out_w": state_out_cond[2],
                    "h_l_out_cold": state_cold_out[2],
                    "h_l_out_w": state_out_evap[2],
                    "T_hl": _STORAGE_T_IN_,
                    "T_lh": _STORAGE_T_IN_,
                    "T_ll": _COLD_STORAGE_T_OUT_,  # 256.0,
                    "Q_dot_h": _Q_DOT_MIN_,
                    "d_temp_min": _D_T_MIN_}

    print(
        f"p-ratio: {state_in_cond[1]/state_out_evap[1]: .2f}, p_low: {state_out_evap[1]/1e5: .2} bar")
    hp0 = cb.hp_simple.HeatPump([myFluid, secFluid, coldFluid], FIXED_POINTS)
    print(hp0.evaluation)
    cop = hp0.calc_heat_pump(verbose=True)
    print(hp0.evaluation)
    fig_ax_act = hp0.hp_plot()
    print(hp0.evaluation, "\n----------------\n")

    out = hp0.evaluation
    print(f"Min and mean dT evaporator: {out['evaporator']['dT-min']}, {out['evaporator']['dT-mean']}")
    print(f"Min and mean dT condenser: {out['condenser']['dT-min']}, {out['condenser']['dT-mean']}")

    print(
        f"COP: {cop},p-ratio: {out['p_high']/out['p_low']:.2f}, p_low {out['p_low']/1e5:.2f} bar")

    print(
        f'exergy loss rate: {out["exergy_loss_rate"]}, eff: {1-out["exergy_loss_rate"]/out["Power"]:.4f}')

    #
    my_dict = cb.hp_simple.read_hp_results(_RESULTS_ +
                                           "\\last_T_H_dot_plot_hp_evaluation_dict.npy")
    T_MAX = hp0.all_states[1][-1][0]  # actual highest storage temperature
    T_MIN = hp0.all_states[3][-1][0]  # actual minimum storage temperature

    # ----------------------ORC ----------------

    # FLUID = "Propane * Butane * Pentane * Hexane"
    # comp = [.75, 0.05, 0.15, 0.05]
    # comp = [0.4,	0.3,	0.3, 0.0]  # [0.164,.3330,.50300,0.0]

    # FLS = "Water"  # Storage fluid
    # FLCOLD = "Methanol"  # Storage fluid for low T
    FLENV = "Water"  #

    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])

    coldFlm = cb.fprop.FluidModel(FLCOLD)
    coldFluid = cb.fprop.Fluid(coldFlm, [1.])

    envFlm = cb.fprop.FluidModel(FLENV)
    envFluid = cb.fprop.Fluid(secFlm, [1.])

    # Condenser(c) and storage (s), secondary fluids fix all, temperatures(T in K),
    # pressures (p in Pa)
    _ETA_S_ = 0.7  # interesting when changed from 0.69 to 0.65, the efficiency
    # decreases, the reason is the low quality along throtteling then
    _ETA_S_P_ = 0.6  # pump
    _STORAGE_T_OUT_ = cb._T_SURROUNDING
    _COLD_STORAGE_T_OUT_ = cb._T_SURROUNDING
    _ENV_T_IN_ = cb._T_SURROUNDING
    _ENV_T_OUT_ = cb._T_SURROUNDING + 5.
    _STORAGE_T_IN_ = T_MAX  # 395.0
    _COLD_STORAGE_T_IN_ = T_MIN
    _STORAGE_P_IN_ = 5e5
    _COLD_STORAGE_P_IN_ = 5e5
    _ENV_P_IN_ = 5e5
    _Q_DOT_MIN_FACTOR = 1.  # and heat_flow rate (W)
    _D_T_SUPER_ = 5  # super heating of working fluid
    _D_T_MIN_ = 2.  # minimum approach temperature (pinch point)
    _COP_CHARGING = 2.925  # needed to calculate Q_env_discharging
    _T_REDUCTION_EVAP = -15  # if the curves cross in the evaporator this parameter may help
    _T_INCREASE_COND = 17.5
    # environment for heat transfer
    state_env_out = envFluid.set_state([_ENV_T_OUT_, _ENV_P_IN_], "TP")
    state_env_in = envFluid.set_state([_ENV_T_IN_, _ENV_P_IN_], "TP")

    # high T-storages
    state_sec_out = secFluid.set_state([_STORAGE_T_OUT_, _STORAGE_P_IN_], "TP")
    state_sec_in = secFluid.set_state([_STORAGE_T_IN_, _STORAGE_P_IN_], "TP")

    #  low T sorages:
    state_cold_out = coldFluid.set_state(
        [_COLD_STORAGE_T_OUT_, _COLD_STORAGE_P_IN_], "TP")
    state_cold_in = coldFluid.set_state(
        [_COLD_STORAGE_T_IN_, _COLD_STORAGE_P_IN_], "TP")

    # working fluid

    state_satv_evap = myFluid.set_state(
        [_STORAGE_T_IN_-_D_T_MIN_-_D_T_SUPER_+_T_REDUCTION_EVAP, 1.], "TQ")  # find high pressure
    p_high = state_satv_evap[1]

    T_OUT = _STORAGE_T_IN_ - _D_T_MIN_
    # Evaporator input comes from the pump-output

    state_out_evap = myFluid.set_state([p_high,
                                        T_OUT], "PT")
    # low pressure, condenser # BA 2023-11-14 three points needed: low, environment and slightly higher
    T_SATL = _COLD_STORAGE_T_IN_ + _D_T_MIN_ + \
        _T_INCREASE_COND  # BA 2024-05-21 be careful
    state_out_cond = myFluid.set_state([T_SATL, 0.], "TQ")  # find low pressure
    p_low = state_out_cond[1]
    # BA changed 2023-12-13 the fixed starting point for the cycle is the  fluid
    # state before the pump now.

    # the other states in the condenser are fixed by the expander outlet and
    # the Q_total : Q_low_stored ratio    eventually p_low must be varied until
    # the balance is fulfilled, since m_dot_w is fixed by the evaporator!

    FIXED_POINTS_ORC = {"eta_s": _ETA_S_,  # expander
                        "eta_s_p": _ETA_S_P_,  # pump
                        "p_low": p_low,
                        "p_high": p_high,
                        "T_hh": _STORAGE_T_IN_,
                        "h_h_out_sec": state_sec_out[2],
                        "h_h_out_w": state_out_evap[2],
                        "h_l_out_cold": state_cold_out[2],
                        "h_l_out_w": state_out_cond[2],
                        "h_env_in": state_env_in[2],
                        "h_env_out": state_env_out[2],
                        "T_hl": _STORAGE_T_OUT_,
                        "T_lh": _COLD_STORAGE_T_OUT_,
                        "T_ll": _COLD_STORAGE_T_IN_,  # 256.0,
                        "Q_dot_h": _Q_DOT_MIN_ * _Q_DOT_MIN_FACTOR,
                        "d_temp_min": _D_T_MIN_,
                        "cop_charging": cop  # needed to calculate Q_env_discharging
                        }

    orc0 = cb.orc_simple.OrganicRankineCycle(
        [myFluid, secFluid, envFluid, coldFluid], FIXED_POINTS_ORC)
    eta_dis = orc0.calc_orc()
    print(f"eta(ORC): {eta_dis:.3f}, COP(HP): {cop:.3f}")
    orc0.hp_plot(fig_ax=fig_ax_act)
    print(f"RTE: {cop*eta_dis:.3f}")
    print(f"warnings: HP {hp0.warning}; ORC {orc0.warning}")



# -*- coding: utf-8 -*-
"""
Run script to evaluate the fluid screening, together with the compressor model

which are stored in two csv-files. First they are sorted then concatenated.
Afterwards, the Paeroto optimal values are selected, plotted and stored.


Created on Mon Feb 12 17:01:22 2024

@author: atakan
"""

from paretoset import paretoset
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sbn
import carbatpy as cb


RUN_NO = 1

f_add ="-sort-rlt-4val" # change this name to have unique file names for each run
directory = r"C:\Users\atakan\sciebo\results\optimal_hp_fluid\fluid_select_restricted\2024-02-12-16-55-ProEthPenBut"

filename1 = directory + r"\\2024-02-12-16-55-ProEthPenBut.csv"
pareto_file = filename1.split(".")[0]+'-sort-evaluated.csv'

f_name = filename1.split(".")[0]+f_add
filename2 = directory + \
    r"\\2024-02-12-16-55-ProEthPenBut-compressor-Roskosch.csv" # the compressor data

data_screen = pd.read_csv(filename1)
keys=data_screen.columns
fl_all =cb. property_eval_mixture.get_fluid(data_screen)
fl_col = fl_all[1]
data =data_screen.sort_values(axis=0, by=fl_col).reset_index(drop=True)
# .reset_index is important, without it, sorting gets lost after concat


eval_dat = pd.read_csv(filename2)
keys2 =eval_dat.columns
eval_sorted=eval_dat.sort_values(axis=0, by=fl_col).reset_index(drop=True)
all_keys=[*keys, *keys2]

combined_data = pd.concat([data, eval_sorted], axis=1, ignore_index=True)
combined_data.columns=all_keys
combined_data.to_csv(f_name+"-combi-new.csv")


what_act = {"x": 'spec_Volume_sup', "y": 'COP_is80', "hue": 'p_ratio',#',# 'COP_comp', #'T_glide_h',
             "size":  'spec_Volume_sup', 'style': 'Temperature_hplt'}

if RUN_NO ==0:
    
##############################################
   
    SUCCESS = cb.property_eval_mixture.data_plot(filename1, what_act)

############################################
    fluids_act, fluid_col_act, fluid_str_act = cb.property_eval_mixture.get_fluid(combined_data)

#########################################

    evaluated_data = cb.property_eval_mixture.eval_is_eff_roskosch(combined_data,
                                      filename1.split(".")[0]+'-sort-evaluated.csv')

###################################
if RUN_NO >=0: # Pareto-Optimal values are searched, for these objectives:
    objectives_act = [ 'T_glide_h',  'eff_sec_law_r_lowT', 'spec_Volume_sup','p_ratio']  #'eff_sec_law_r'] # 'spec. Volume_sup',
    sense_act = ["max", "max", "min", "min"]
    
    what_act["x"]=objectives_act[0]
    what_act["y"]=objectives_act[1]
    
    optimal_data = cb.optimize.pareto(pareto_file, [objectives_act,sense_act])
    opt_values = optimal_data["all_values"][ optimal_data["optimal_mask"]]
    opt_values.to_csv(f_name+"-pareto.csv")
    graph = sbn.relplot(data=opt_values.round(3),
                        y=what_act["y"],
                        hue=what_act["hue"],
                        size=what_act["size"],
                        
                        x=what_act["x"])
    graph.savefig(f_name+"-plot.jpg", dpi =300)
    with open(f_name+"-objectives-.txt", "w", encoding="utf-8") as file:
              file.write(str(list(zip(objectives_act,sense_act))))
              
    
    graph2 = sbn.relplot(data=optimal_data["all_values"],
                        y="is_eff",
                        hue="p_ve",
                        #size="eff_sec_law_r_lowT",
                        
                        x='spec_Volume_sup')
    graph2.savefig(f_name+"-plot_is_eff.jpg", dpi =300)
    # val = cb.property_eval_mixture.get_cycle_points(opt_values, 80)
    # val2 = cb.property_eval_mixture.get_cycle_points(opt_values, 104)
    # figure, axis =plt.subplots(1,1)
    # axis.plot(val[2,:]- val[2,:].min(),val[0,:],"o:k")
    # axis.plot(val2[2,:]- val2[2,:].min(),val2[0,:],"v:b")
    



# -*- coding: utf-8 -*-
"""
find the two points with the (nearly) minimum distance between two curves
Created on Tue Nov 14 13:19:00 2023

@author: atakan
"""

import numpy as np
import matplotlib.pyplot as plt



x = np.linspace(0, 12)
shift =1
line = -1.9*x + 5+shift
second = 5*np.sin(x) + 5 * x + 4

f,ax = plt.subplots(1,1)
ax.plot(x,line)
ax.plot(x,second)
difference =line-second
#difference[25] =1
ax.plot(x,difference,"v")
sum_shifted = np.sum(np.abs(difference))
grad = np.gradient(difference,x)
grad2 = np.gradient(grad,x)
ax.plot(x,grad,":")
ax.plot(x,grad2,"o")
idx = np.where(np.sign(grad[:-1]) != np.sign(grad[1:]))[0] + 1
idx_min =np.where(grad2[idx]>0)
x_min =x[idx[idx_min]]
print(idx, difference[idx], grad[idx],grad2[idx], x_min)
# if len(idx) > 0:
#     wanted = min(d_temps[idx] - self.d_temp_separation_min)
# else:
#     wanted = mind_temp-self.d_temp_separation_min



"""Top-level package for carbatpy."""

__author__ = """Burak Atakan"""
__email__ = 'burak.atakan@uni-due.de'


# __all__ =[]
import os
import sys
# read version from installed package
from importlib.metadata import version

import pandas as pd


__version__ = version('carbatpy')

pd.set_option("mode.copy_on_write", True)

sys.path.insert(0,os.path.abspath('../src'))
sys.path.insert(0,os.path.abspath('..'))


try:
    from cb_config import _T_SURROUNDING, _RESULTS_DIR, \
                            _P_SURROUNDING,_CARBATPY_BASE_DIR, _TREND
   
except:
    from .cb_config import _T_SURROUNDING, _RESULTS_DIR, \
                            _P_SURROUNDING,_CARBATPY_BASE_DIR, _TREND
 
import carbatpy.models as models
import carbatpy.utils as utils
import carbatpy.helpers as helpers



from carbatpy.utils import exergy_loss as exlo
from carbatpy.helpers import file_copy

import carbatpy.models.fluids.fluid_props as fprop

from carbatpy.models.components import compressor_simple, throttle_simple
from carbatpy.models.components import heat_exchanger_thermo_v2 as hex_th

from carbatpy.models.coupled import heat_pump_simple_v2 as hp_simple
from carbatpy.models.coupled import orc_simple_v2 as orc_simple
from carbatpy.models.coupled import read_cycle_structure
from carbatpy.utils import curve_min_distance_finder, property_eval_mixture, optimize




# -*- coding: utf-8 -*-
"""
Created on Sun Nov  5 10:01:28 2023

@author: atakan
"""
import numpy as np
import carbatpy as cb

n_pos=[0,2,4]
def s_irr_flow(states, mass_flow_rates, verbose=False, entropy_position=4):
    """
    Irreversible entropy production rate for two fluid flows or an adiabatic
    heat exchanger. It is intended to use for heat pumps, where energy is
    stored at low and at high temperature.

    Parameters
    ----------
    states : numpy-array [n,2,7]
        the states to analyse ordered as entering fluids[:,0,:] and exiting
        fluid [:,1,:] the states are as defined in fluid_props, with the 4th
        value being the specific entropy in J/(kg K)(see below).
    mass_flow_rates : numpy array of length n
        the mass flow rates in SI units (kg/s).
    verbose : Boolean, optional
        should we print? The default is False.
    entropy_position : Integer, optional
        in the states array, where to find thge specific entropy.
        The default is 4.

    Returns
    -------
    float
        entropy production rate in W/K.

    """

    # state in always first
    s_irr = -(states[:, 0, entropy_position] -
              states[:, 1, entropy_position]) * mass_flow_rates
    if verbose:
        print("Sirr:", s_irr, states[:,:,n_pos], mass_flow_rates, s_irr.sum())
    return s_irr.sum()


def exergy_loss_alternative(states, mass_flow_rates,  power_in=0.,
                            temp_surrounding=cb._T_SURROUNDING,
                            verbose=False, entropy_position=4,
                            enthaply_position=2):
    """
    assumption both fluids enter in equilibrium with the surrounding
    if entropy is reduced, T is below the environment T and exergy is positive
    
    but this not correct yet, should be implemented correctly 
    """
    ds = states[:, 0, entropy_position] - states[:, 1, entropy_position]
    dh = states[:, 0, enthaply_position] - states[:, 1, enthaply_position]
    dex = (dh - ds * temp_surrounding) * mass_flow_rates
    if verbose:
        print("Ex-Alt:", dh*mass_flow_rates, ds*mass_flow_rates, mass_flow_rates, dex, power_in)
        print("in:",states[:,0,n_pos])
        print("out:",states[:,1,n_pos])
        print("dex",dex)
    dex_total = power_in + np.sum(dex)
    return dex_total  # must be checked, if more than two streams are analyzed


def exergy_loss_flow(states, mass_flow_rates, power_in=0.,
                     temp_surrounding=cb._T_SURROUNDING,
                     verbose=False, entropy_position=4):
    """
    Calculate the exergy loss rate for two fluid flows or an adiabatic
    heat exchanger. It is intended to use for heat pumps, where energy is
    stored at low and at high temperature. Uses s_irr_flow. All parameters are
    described there, except one. Gouy-Stodola


    Parameters
    ----------
    states : TYPE
        DESCRIPTION.
    mass_flow_rates : TYPE
        DESCRIPTION.
    power_in : TYPE, optional
        DESCRIPTION. The default is 0..
    temp_surrounding : Float, optional
        surrounding temperature in K. The default is cb._T_SURROUNDING.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.
    entropy_position : TYPE, optional
        DESCRIPTION. The default is 4.

    Returns
    -------
    float
        Exergy loss rate in W.

    """

    s_irr = s_irr_flow(states, mass_flow_rates,
                       verbose=verbose, entropy_position=entropy_position
                       )

    return s_irr * temp_surrounding #+ power_in


if __name__ == "__main__":
    FLUID = "Propane * Pentane"
    comp = [.50, 0.5]
    P_OUT = 5e5
    power = 15000.00 
    m_dots = np.array((1.4, 1))  # important different signs
    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)
    
    state_in = myFluid.set_state([320., 19e5], "TP")
    state_out = myFluid.set_state([300, 19e5], "TP")
    
    state_in_2 = myFluid.set_state([290., 19e5], "TP")
    h2o = state_in_2[2] + (m_dots[0]*(state_in-state_out)[2] + power)/m_dots[1]
    state_out_2 = myFluid.set_state([h2o, 19e5], "HP")
    print("States\n",state_in[n_pos], state_out[n_pos],
          state_in_2[n_pos], state_out_2[n_pos], h2o, m_dots)

    states = np.array([[state_in, state_out], [state_in_2, state_out_2]])
    print("Irr-Gouy-Stod1\n",s_irr_flow(states, m_dots), exergy_loss_flow(states, m_dots, power_in=power))
    print("Ex_loss2\n", exergy_loss_alternative(states, m_dots, power_in=power))



# -*- coding: utf-8 -*-
"""
Started a class for flow devices, especially mixing chambers, but could also be
used for heat exchangers or machines?

at the moment, the only device is for adiabtaic mixing, but also of
fluid mixtures of different compounds and composition.
ouputs as dictionary (state variables to be fixed?)
inputs=list of states (Fluid)
mix = list of 2 lists [[input 1(number), input 2 (number)], [output (number? or different?)]]

Created on Fri May 31 15:16:09 2024

@author: atakan
"""


import copy

# import src.models.fluids.fluid_props as fprop
import numpy as np
from scipy.optimize import minimize  # root, root_scalar
import matplotlib.pyplot as plt
import carbatpy as cb


class FlowDevice:

    def __init__(self, inputs, outputs, flow_rates_in, mix=[],
                 name="flow-device",
                 **calc_parameters):
        """
        Flow Device = mixing chamber

        At the moment the species, the mass balance, and the energy balance
        for an adiabatic chanber with given output pressure is calculated. Two
        fluids entering, at the moment only one is exiting.

        Parameters
        ----------
        inputs : list of Fluids
            the (2) fluids entering with the correct states.
        outputs : List of Fluids
            the (actually single) fluid state exiting.
        flow_rates_in : list
            the mass flow rates in kg/s.
        mix : List of two lists, optional
            In the first list the input states are numbered, in the second the
            output states example: [[0,1],[0]]. The default is [].
        name : string, optional
            the name of the device. The default is "flow-device".
        **calc_parameters : dictionary
            example {"calc_type": "HP", "H": "const", "p_out": 1e5}, at the
            moment only p_out values should be changed.

        Returns
        -------
        None.

        """
        self.inputs = inputs
        self.outputs = outputs
        self.flow_rates_in = flow_rates_in
        self.name = name
        # print(mix)
        match len(mix):
            case 0:
                self.flow_rates_out = flow_rates_in
            case 2:
                self.molecular_weights = []
                self.mix_species = {}
                spec_info = []
                self.n_dot_total = 0
                names = []
                m_flows_list = []
                self.flow_rate_out = [np.array(flow_rates_in).sum()]
                enthalpy_total = []

                for i_in in mix[:][0]:
                    act_info = []
                    state_act = inputs[i_in].properties.state
                    state_act = inputs[i_in].set_state([state_act[1],
                                                        state_act[2]], "PH",
                                                       cb.fprop._TRANS_STRING)
                    molecular_weight = inputs[i_in].properties.molecular_mass
                    n_dot = molecular_weight * flow_rates_in[i_in]
                    self.n_dot_total += n_dot
                    names.extend(inputs[i_in].fluidmodel.fluid_to_list())
                    mole_frac = inputs[i_in].composition
                    mol_flows = np.array(mole_frac) * n_dot
                    m_flows_list.extend(list(mol_flows))
                    enthalpy_total.append(state_act[2] * n_dot)

                all_species = dict.fromkeys(names, 0)

                m_flows = np.array(m_flows_list) / self.n_dot_total

                enthalpy_out = (np.array(enthalpy_total) /
                                self.n_dot_total).sum()

                # print(m_flows.sum(), m_flows, enthalpy_out)

                for i_spec, spec in enumerate(names):
                    all_species[spec] += m_flows[i_spec]
                # all_species.extend(inputs[i_in].fluidmodel.fluid_to_list())
                self.all_species = all_species
                rp_names = "*".join(list(self.all_species.keys()))
                st_mix = cb.fprop.init_fluid(rp_names, all_species.values())
                # st_mix= cb.fprop.Fluid(fl_mix,
                #                        list(my_flow_dev.all_species.values()))

                self.outputs[mix[1][0]] = st_mix
                match calc_parameters["calc_type"]:
                    case "HP":
                        if calc_parameters["H"] == "const":
                            self.st_out = st_mix.set_state([enthalpy_out,
                                                            calc_parameters["p_out"]],
                                                           calc_parameters["calc_type"])
                    case _:
                        self.warning = 998
                        print(
                            f"FlowDevice: calcukation type not implemented: {calc_parameters['calc_type']}")

            case _:
                self.warning = 999
                print(f"FlowDevice: not implemented for length of mix: {mix}")

        #


if __name__ == "__main__":
    FLUID = "Propane * Pentane"
    comp = [.50, 0.5]
    myFluid = cb.fprop.init_fluid(FLUID, comp)
    calc_parameters_act = {"calc_type": "HP", "H": "const", "p_out": 1e5}

    st0 = myFluid.set_state([300., 1e5], "TP")
    FLUID2 = "Propane * Ethane"
    comp2 = [.50, 0.5]
    secFluid = cb.fprop.init_fluid(FLUID2, comp2)
    st0 = secFluid.set_state([400., 1e5], "TP")
    inputs_act = [myFluid, secFluid]
    outputs_act = [myFluid]
    mix_act = [[0, 1], [0]]
    flows_act = [.01, .02]
    my_flow_dev = FlowDevice(inputs_act, outputs_act, flows_act, mix_act,
                             **calc_parameters_act)
    print(my_flow_dev.all_species, my_flow_dev.st_out)



# -*- coding: utf-8 -*-
"""
Comparison of cost estimation taken from two books for reciprocating
compressors

cost2: chapter 21 of  Chemical engineering design : principles, practice and 
economics of plant and process design
Gavin Towler, Ray Sinnott. 3rd ed.; 2022

cost_function: Chemical Process Equipment Selection and Design
Third Edition
James R. Couper
W. Roy Penney
James R. Fair
Stanley M. Walas

"""


def cost_function(size, what):
    all_data = {"Compressor-Reciprocating": [2.6e5, 2700, 0.75, 93, 16800]}
    constant, multi, exponent, min_size, max_size = all_data[what]
    warning = 1
    if size > min_size and size < max_size:
        warning = 0
    cost = constant + multi * size**exponent
    return round(cost,0), warning


if __name__ == "__main__":
    size_act = 100  # kW
    what_act = "Compressor-Reciprocating"
    print(
        f"Prize of {what_act} at size {size_act}: {cost_function(size_act, what_act)}")
    horse_power = 0.735  # kW
    cost2 = round(7.19 * (size_act / horse_power)**.61 *1000, 0)
    print(f"second prize: {cost2}")



# -*- coding: utf-8 -*-
"""
Created on Sun May 21 08:51:33 2023

@author: atakan
"""


import copy
import carbatpy as cb
# import src.models.fluids.fluid_props as fprop
import numpy as np
from scipy.optimize import root
import matplotlib.pyplot as plt


class StaticHeatExchanger:
    """
    Class for static counter-flow heat exchanger

    means: no time dependence and no heat transfer coefficients * areas
    are used (UA)! Instead a minimum approach temperature is tried to be met.
    At the moment, this is mainly done by varying one of the mass flow rates.
    But this is sometimes not enough and a variation of the working fluid
    pressure will soon be included.
    Only the first law and second law will be checked (the latter must be
    improved).

    """

    def __init__(self, fluids, h_dot_min, temp_out_s,
                 points=50, d_temp_separation_min=0.5, calc_type="const",
                 name="evaporator"):
        """
        class to calculate (static/steady state) heat-exchangers

        includes pinch-point analysis and plotting,
        only implemented for simple thermodynamic calculations
        (no convection coefficients and heat exchanger areas regarded yet)

        Parameters
        ----------
        fluids : list of 2 fprop.Fluid
            The definition of the two fluids, as they enter the heat exchanger.
        h_dot_min : float
            enthalpy flow rate (W) which has to be transfered.
        temp_out_s : float
            exit temperature of the secondary fluid.
        points : integer, optional
            for how many points (array) shall the minimum approach temperature
            be checked and properties be returned (for plotting etc.)
            The default is 30.
        d_temp_separation_min : float, optional
            Minimium approach temperature (pinch point) between the two fluids.
            The default is 0.5.
        calc_type : string, optional
            which calculation type shall be performed; only one implemented so
            far. The default is "const".
        name : string, optional
            name of the heat exchanger. The default is "evaporator".

        Returns
        -------
        None.

        """
        self.fluids = fluids
        self.q_dot = h_dot_min
        self.m_dot_s = 0
        self.temp_out_s = temp_out_s
        self.points = points
        self. d_temp_separation_min = d_temp_separation_min
        self.heating = False
        if temp_out_s > fluids[1].properties.temperature:
            self.heating = True  # condenser
        self.calc_type = calc_type
        self.name = name
        self.all_states = np.zeros((points, len(cb.fprop._THERMO_STRING)))
        self.h_in_out = np.zeros((2, 4))
        self.warning = 0


    def pinch_calc(self, m_dot_w_factor=1, verbose=False):
        """
        Calculate the changes in enthalpy and temperature in the heat exchanger

        counter-flow hex assumed! Both flows are isobaric.
        Is used to check, whether the second law is violated. The factor can
        be used to vary the mass flow rate of the working fluid, until no
        violation is found (done in root finding).

        Parameters
        ----------
        m_dot_w_factor : float, optional
            factor to divide the working fluid mass flow rate, as derived from
            the enrgy balance, in order to avoid a crossing of temperature
            curves. The default is 1.

        Raises
        ------
        Exception
            if temperatures are not consistent.

        Returns
        -------
        m_dot_w : float
            working fluid mass flow rate (kg/s.
        d_tempall : numpy-array
            The temperature differences along the counter-flow heat exchanger.
        w_array : array
            properties of the working fluid along the heat exchanger
            (T,p,h, etc. see fluid class).
        s_array : array
            properties of the secondary fluid along the heat exchanger
            (T,p,h, etc. see fluid class).

        """
        w_in = self.fluids[0]
        s_in = self.fluids[1]

        w_out = copy.copy(self.fluids[0])
        s_out = copy.copy(self.fluids[1])

        self. h_in_out[1, 0] = h_in_s = s_in.properties.enthalpy
        h_in_w = w_in.properties.enthalpy

        # fixed output temperature, secondary fluid
        s_out.set_state([self.temp_out_s, s_in.properties.pressure], "TP")
        if self.heating:
            condenser = 1
        else:
            condenser = -1

        self. h_in_out[1, 1] = h_dot_s = (
            s_out.properties.enthalpy - s_in.properties.enthalpy) * condenser
        # fixed heat flow,  determines mass flow rate
        self.m_dot_s = self.q_dot / h_dot_s

        outw = w_out.set_state([s_in.properties.temperature +
                                self.d_temp_separation_min * condenser,
                                w_in.properties.pressure], "TP")
        hw_max = outw[2]
        h_dot_w_max = np.abs(w_in.properties.enthalpy - hw_max)
        mw_max = self.q_dot / h_dot_w_max
        # bei reiner propan WP geht etwas schief, BA
        if verbose:
            print("mw_max", mw_max, self.q_dot)
        m_dot_w = mw_max * m_dot_w_factor  # scaling for minimum approach T
        h_dot_w = h_dot_w_max / m_dot_w_factor
        h_out_w = h_in_w - condenser * h_dot_w
        if np.isscalar(h_dot_w):
            pass
        else:
            h_dot_w = h_dot_w[0]
            h_out_w = h_out_w[0]

        self. h_in_out[0, 0] = h_out_w

        self. h_in_out[0, 1] = h_dot_w

        # check pinch, first secondary fluid
        h_array = np.linspace(h_in_s, h_in_s + h_dot_s *
                              condenser, self.points)
        values = np.zeros((self.points, 2))
        values[:, 0] = h_array
        values[:, 1] = s_out.properties.pressure

        s_array = s_out.set_state_v(values, given="HP")

        # now working fluid:
        h_array = np.linspace(self.h_in_out[0, 0],
                              self.h_in_out[0, 0] +
                              self.h_in_out[0, 1] * condenser,
                              self.points)

        values = np.zeros((self.points, 2))
        values[:, 0] = h_array.T
        values[:, 1] = w_out.properties.pressure

        w_array = w_out.set_state_v(values, "HP")
        d_tempall = w_array[:, 0]-s_array[:, 0]
        if verbose:
            if condenser == -1:
                print("evap", d_tempall[0], d_tempall[-1],
                      d_tempall.max(), d_tempall.min())
            else:
                print("cond", d_tempall[0], d_tempall[-1],
                      d_tempall.min(), d_tempall.max())
        return m_dot_w, d_tempall, w_array, s_array

    def pinch_root(self, factor, verbose=False):
        """
        function for root-finding of the minimum approach temperature

        a factor to divide the working fluid mass flow rate is varied. input
        for root

        Parameters
        ----------
        factor : float
            as said above.

        Returns
        -------
        float
            root tries to reach a value of 0.

        """
        mw, d_temps, w, s = self.pinch_calc(factor)
        mind_temp = d_temps.min()
        maxd_temp = d_temps.max()
        if mind_temp/maxd_temp <= 0:
            return 1e6
        # todo: check that all signs of d_temps are equal, if not:return values !=0
        if self.heating:
            if verbose:
                print("H", mind_temp - self.d_temp_separation_min,
                      d_temps.min(), self.d_temp_separation_min, mw)
            return mind_temp - self.d_temp_separation_min
        else:
            if verbose:
                print("HC", maxd_temp + self.d_temp_separation_min,
                      d_temps.max(), self.d_temp_separation_min, mw)
            return maxd_temp + self.d_temp_separation_min

    def find_pinch(self):
        """
        Function tries to vary one mass flow rate until a minimum approach
        temperature is reached. If this is not successful, the working fluid
        pressure is varied within 20 percent. This is then also the new pressure
        within the heat exchanger. If this is also not succesful,
        self.warning is set to 1. This should be checked.

        Returns
        -------
        float
            the factor to multiply the mass flow rate with, to reach the 
            minimum approach temperature. Igf not succesful the initial factor
            is returned.

        """
        verbose = False
        w_initial = copy.copy(self.fluids[0])
        p0 = w_initial.properties.pressure
        enthalpy = w_initial.properties.enthalpy

        P_VARIATION = 0.2
        N_VARIATION = 21
        if self.heating:
            start = 1.05
            p_var = np.linspace(p0, p0*(1+P_VARIATION), N_VARIATION)
        else:
            start = .95
            p_var = np.linspace(p0, p0*(1-P_VARIATION), N_VARIATION)
        try:
            for pressure in p_var:
                self.fluids[0].set_state([pressure, enthalpy], "PH")
                result = root(self.pinch_root, start)
                if verbose:
                    print(
                        f"pressure:{pressure}, result {result}, heating {self.heating}")

                if result.success:
                    return result.x[0]

        # except:
        except Exception as inst:
            print(type(inst))    # the exception type
            print(inst.args)     # arguments stored in .args
            print(inst)          # __str__ allows args to be printed directly,
            # but may be overridden in exception subclasses

            result = np.array([1e4])
            print("root-exception", self.heating)
            return result
        print("root-finding problem! (in heat_exchanger_thermo_v2.find_pinch)",
              result)
        print(f"Heating: {self.heating}")
        self.warning = 1
        return self.factor

    def pinch_plot(self, factor, plot_fname="", plotting=True):
        print(f"------pinch-plot running -----plot:{plotting}")
        m_dot_w, d_tempall, w_array, s_array = self.pinch_calc(factor)

        h_w_plot_array = (
            w_array[:, 2] - self.h_in_out[0, 0]) * m_dot_w / self.m_dot_s
        if plotting:
            fig, ax = plt.subplots(1, 1)
            ax.plot(s_array[:, 2] - self.h_in_out[1, 0], s_array[:, 0], "v")
            ax.plot(h_w_plot_array, w_array[:, 0], "o")
            ax.set_xlabel(
                "specific enthalpy flow per mass of secondary fluid / (J / kg)")
            ax.set_ylabel("temperature / (K)")
            ax.set_title("heat exchanger, simple")
        if plot_fname != "":
            fig.savefig(plot_fname)

        return m_dot_w, d_tempall, w_array, s_array


if __name__ == "__main__":
    # two test cases condenser and evaporator:
    FLUID = "Propane * Pentane"  # working fluid
    FLS = "Methanol"  # "Water"  # secondary fluid
    comp = [.50, 0.5]
    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])

    # Condenser, secondary fluid fixes all:
    sT_in = 300.0
    sT_out = 370.0
    sp_in = 5e5
    h_dot_min = 1e3

    state_sec_in = secFluid.set_state([sT_in, sp_in], "TP")
    # working fluid
    d_temp_super = 5.
    T_dew = sT_out
    state_in = myFluid.set_state([T_dew, 1.], "TQ")  # find minimum pressure

    T_in = T_dew + d_temp_super
    state_in = myFluid.set_state([myFluid.properties.pressure,
                                  myFluid.properties.temperature + d_temp_super],
                                 "PT")
    d_temp_min = 3.0

    hex0 = StaticHeatExchanger([myFluid, secFluid], h_dot_min, sT_out,
                               d_temp_separation_min=d_temp_min)
    mw, d_tempall, w, s = hex0.pinch_calc(1.1)
    factor0 = hex0.find_pinch()
    mw0, d_tempall0, w0, s0 = hex0.pinch_plot(factor0)
    print(f"w0 {w0[0]}")

    # Evaporator: ----------------------------

    sT_in = 300.0
    sT_out = 276
    sp_in = 5e5
    h_dot_min = 1e3

    state_sec_in = secFluid.set_state([sT_in, sp_in], "TP")
    d_temp_min = 2.0
    T_in = sT_out  # - d_temp_min
    state_in = myFluid.set_state([T_in, .1], "TQ")
    print("state in", state_in)

    hex0 = StaticHeatExchanger([myFluid, secFluid], h_dot_min, sT_out,
                               d_temp_separation_min=d_temp_min)
    mw, d_tempall, w, s = hex0.pinch_calc(1)
    factor = hex0.find_pinch()
    mw, d_tempall, w, s = hex0.pinch_plot(factor)
    print(f"w {w[0]}")

    p_out = 5e5

    # state_out = throttle(p_out, myFluid)
    # print("Throttle:\nInput:", state_in,"\nOutput:",
    # state_out,"\nDifference", state_out-state_in)



"""
Demo for using xgboost with sklearn
===================================
"""
import multiprocessing
import pandas as pd
import matplotlib.pyplot as plt

from sklearn.model_selection import train_test_split
from sklearn.model_selection import GridSearchCV

import xgboost as xgb

if __name__ == "__main__":
    which_all = range(93,101)
    which_part = [93,97,99,100]
    which =which_all
    directory = r"C:\Users\atakan\sciebo\results\optimal_hp_fluid\fluid_select_restricted\2024-02-12-16-55-ProEthPenBut"

    filename1 = directory + r"\\2024-02-12-16-55-ProEthPenBut.csv"
    data_file = filename1.split(".")[0]+'-sort-evaluated.csv'
    data =pd.read_csv(data_file)
    good = data["is_eff"]<1   # =1 -> error/two phase
    data = data[good]
    
    col_act =data.columns
    y =data[["is_eff", "degree_delivery"]]
    x_names =col_act[which] #  [93:102] # columns with mole fractions, pressure etc.
    X = data[x_names]
    X_train, X_test, y_train, y_test = train_test_split( X, y,
                                                        test_size=0.3,
                                                        random_state=21)
    print("Parallel Parameter optimization")
    # X, y = fetch_california_housing(return_X_y=True)
    xgb_model = xgb.XGBRegressor(
        n_jobs=multiprocessing.cpu_count() // 2, tree_method="hist"
    )
    clf = GridSearchCV(
        xgb_model,
        {"max_depth": [2, 4, 6], "n_estimators": [50, 100, 200, 400],
         'eta':[.1,.3,.6], "gamma":[0,2,20]},
        verbose=1,
        n_jobs=2,
    )
    for objective in ["is_eff", "degree_delivery"]:
        print(f"---------\n {objective} \n")
        clf.fit(X_train, y_train[objective])
        best =clf.best_estimator_
        
        imp =list( zip(x_names, best.feature_importances_.round(4)))
        print(clf.best_score_)
        print(clf.best_params_)
        print(imp)
        plt.figure(objective)
        plt.plot(y_train[objective], best.predict(X_train), "xk")
        plt.plot(y_test[objective], best.predict(X_test), "ob")



# -*- coding: utf-8 -*-
"""
Created on Sun May 21 08:51:33 2023

@author: atakan
"""


import carbatpy as cb
import numpy as np


def throttle(p_out, fluid, m_dot=1.0, calc_type="const_h", name="throttle",
             plot_info={}):
    """
    throttle output state calculation

    so far only for a constant enthalpy

    Parameters
    ----------
    p_out : float
        output pressure.
    fluid : fprop.Fluid
        entering fluid, including properties, composition, and model.
    m_dot : float, optional
        mass flow rate (in kg/s). Default is 1
    calc_type : string, optional
        how to calculate, so far, only one implemented. The default is
        "const_h".
    name : string, optional
        name of the device. The default is "throttle".
    plot_info : dictionary, optional
        if not empty a Figure, an Axes, a list of What shall be plotted,
        a list with the colour/styles and a list with the labels must be 
        passed. in "what", the two numbers coincide with the fluid THERMO
        order. The x-shift can be used in cycle calculations, to shift the
        curves, by the value (it will be added).
        The names in the dictionary are: "fig", "ax", "what","col",
        "label", "x-shift". Default is empty.

    Returns
    -------
    state_out : array of float
        compressor output state containing [T,p,h,v,s,q].

    """
    state_in = fluid.properties.state
    if calc_type == "const_h":
        state_out = fluid.set_state([fluid.properties.enthalpy, p_out], "HP")
    else:
        raise Exception(
            f"The option{calc_type} is not yet implemented for throttles")
        
    cb.compressor_simple.plot_temp_h_flow(state_in, state_out, m_dot, plot_info)   
    # 
    return state_out


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    M_DOT = 1e-3
    fig0, ax0 = plt.subplots()
    PLOT_INFO ={"fig":fig0, "ax":ax0, "what":[2,0],"col":["r:","k"],
    "label":["throttle","xx"], "x-shift":[0,0]}
    
    FLUID = "Propane * Pentane"
    comp = [.50, 0.5]
    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)
    state_in = myFluid.set_state([320., 19e5], "TP")

    P_OUT = 5e5

    state_out_main = throttle(P_OUT, myFluid, plot_info=PLOT_INFO)
    print("Throttle:\nInput:", state_in, "\nOutput:",
          state_out_main, "\nDifference", state_out_main-state_in)



# -*- coding: utf-8 -*-
"""
find the straight line which just touches a curve, while having the minimum
area between the curves.

2024-02-11
part of carbatpy

@author: atakan
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize


def straight_diff(param, values, below=True):
    """
    Calculating the integral between a straight line and points from a curve

    the line is defined by the param values [slope,intercept], values[0,:] are
    the x values, values[1,:] are the y-values. below decides, whether the
    straight line should be below or not. This function is used for
    minimization.
    If the curves cross each other or if the line is on the wrong side,
    a large value is returned.

    Parameters
    ----------
    param : list of float
        [slope,intercept].
    values : numpy.array [2,n_points]
        values[0,:] are the x values, values[1,:] are the y-values.
    below : boolean, optional
        is the straight line below. The default is True.

    Returns
    -------
    float
        integral of the difference between the curves.

    """
    difference = values[1, :] - (param[0] * values[0, :] + param[1])

    positive = np.any(difference > 0)
    negative = np.any(difference < 0)

    crossing = (positive > 0 and negative > 0)
    wrong_side = (positive > 0 and not below) or (negative > 0 and below)

    penalize = np.abs(100 * (values[1, :].max() - values[1, :].min())
                      * (values[0, -1] - values[0, 0]))
    if crossing or wrong_side:
        return penalize
    if below and positive > 0:
        integral = np.trapz(difference, values[0])

    else:
        integral = np.trapz(-difference, values[0])
    return integral


def find_min_approach(values_in, below=True, delta=0):
    """
    Finds a straight line which approaches a curve, without crossing.

    Mainly for finding the minimum approach of a fluid with phase change
    to a fluid with a constant heat capacity along heat transfer. One can
    select, whether the straight line should be *below* the curve or above.

    Parameters
    ----------
    values : numpy array [2,n]
        [0,n] are the x-values, [1,n] are the y-values.
    below : boolean, optional
        should the straight line be below? The default is True.
    delta : float, optional
        wanted least distance between curves (absolute value.)

    Returns
    -------
    output : dictionary
        keys are "success", "integral": the integral between curve and line,
        "line_par": slope and intercept of the line (np.array),"below":as above,
        "message": result.message from minimize}.

    """
    values = np.copy(values_in)
    factor = 1
    if below == True:
        factor = -1
    values[1] = values[1] + factor * delta
    para0 = np.polyfit(values[0], values[1], 1)
    min_val = values[1].min()
    max_val = values[1].max()
    diff_val = max_val - min_val
    if below:
        para0[1] -= diff_val
    else:
        para0[1] += diff_val
    result = minimize(straight_diff, para0, args=(values, below),
                      method='Nelder-Mead', tol=1e-4,
                      options={'maxiter': 1e3})
    difference = (values[1] - np.polyval(result.x, values[0])).mean()

    output = {"success": result.success,
              "integral": result.fun,
              "line_par": result.x,
              "below": below,
              "message": result.message,
              "mean_difference": difference - factor * delta}
    return output


def diff_mean(high, low, dist_min=0., optimize=False):
    """
    calculate the mean distance between two equally-spaced arrays

    In thermodynamics/heat exchangers for minimum approach temperature. The size
    of both arrays must be the same. If the curves cross, success will be False!

    Parameters
    ----------
    high : numpy array (length m)
       the high values.
    low : numpy array(length m)
       the low values.
    dist_min : float, optional
        minimum required distnce, must be positive. The default is 0..
    optimize : boolean, optional
        if used for optimization. The default is False.

    Returns
    -------
    dictionary or float
        for optimization the mean difference is returned, else a dictionary
        with success, mean-difference, all differences, dist_min.

    """
    success = True
    if len(high) != len(low):
        print("wrong array lengths")
        return 1e9
    difference = high - low
    mean_diff = difference.mean()
    problem = np.any(difference < dist_min)
    if problem > 0:
        success = False
    if optimize:
        if success:
            return mean_diff
        else:
            return (high.max() - low.min()) * 5
    else:
        return {"success": success,
                "mean_difference": mean_diff,
                "dist_min": dist_min,
                "all_differences": difference}


if __name__ == "__main__":
    # generating test points
    x_val = np.linspace(1, 12)
    y_val = -5 * np.sin(x_val) + 5 * x_val + 4 - 100 / x_val
    values_act = np.array([x_val, y_val]) # input for find_min_approach
    lower = np.polyval([10, -120], values_act[0])

    delta = 5
    ###############################
    # for arbitrary arays, comparison
    mean_difference = diff_mean(values_act[1], lower, 14)
    print(mean_difference)
    
    ###############################
    # finding an optimal straigh line
    solution_below = find_min_approach(values_act, True, delta)
    solution_above = find_min_approach(values_act, False, delta)

    # Plotting ####################
    figure, axis = plt.subplots(1, 1)
    axis.plot(values_act[0], values_act[1], "ko")
    axis.plot(values_act[0], lower,"g:")
    axis.plot(values_act[0], np.polyval(
        solution_below["line_par"], values_act[0]), "b")
    axis.plot(values_act[0], np.polyval(
        solution_above["line_par"], values_act[0]), "r")



# -*- coding: utf-8 -*-
"""
Created on Sun May 21 10:52:09 2023

@author: atakan
"""

import copy
import numpy as np
import matplotlib.pyplot as plt

import carbatpy as cb

_RESULTS_ = r"C:\Users\atakan\sciebo\results\\"


class HeatPump:
    """
    Heat pump class for simple thermodynamic calculations for energy storage

    Two pairwase storages at two different temperatures are assumed,
    always starting at room temperature (around 300K). Thus, three fluids are
    expected: the working fluid, the storage fluid  or sink at high temperature
    (h) and the source storage fluid at low temperature.
    The minimum and maximum temperatures of all
    storags are fixed and the heat flow rate to the high-T storage are
    given. Also, isentropic efficiencies are fixed so far, but they should use
    a function later.
    """

    def __init__(self,  fluids, fixed_points, components=[]):
        """
        Heat pump class for simple thermodynamic calculations for energy storage

        storages at two different temperatures, always starting at room
        temperature (around 300K). The minimum and maximum temperatures aof all
        storags are fixed and the heat flow rate to the high-T storage are
        given. Also, isentropic efficiencies.
        Example for
        fixed_points = {"eta_s": 0.65,
                        "p_low": p_low,
                        "p_high": p_high,
                        "T_hh": _HIGH_T_STORAGE_HIGH_TEMP,
                        "T_hl": _HIGH_T_STORAGE_LOW_TEMP,
                        "T_lh": _LOW_T_STORAGE_HIGH_TEMP,
                        "T_ll": _LOW_T_STORAGE_LOW_TEMP,
                        "Q_dot_h": 3000.0,
                        "d_temp_min": _D_T_MIN_}

        Parameters
        ----------
        fluids : list of three Fluid
            working fluid, secondary fluid to store at high T, cold fluid
            to store at low T.
        fixed_points : Dictionary
            all the fixed points are summarized here, the isentropic efficiency,
            the lower pressure(evaporator), the high and low temperatures of
            the high temperature storages, the same for the low temperature
            storages(source), the heat to be transfered at high T, the minimum
            approach temperature.
        components : TYPE, optional
            DESCRIPTION. The default is [].

        Returns
        -------
        None.

        """
        self.components = components
        self.fixed_points = fixed_points
        self.fluids = fluids
        self.all_states = []
        self.m_dots = []
        self.evaluation = {"Q_dot_h": self.fixed_points["Q_dot_h"],
                           "Power": 0.0,
                           "T_hh": self.fixed_points["T_hh"],
                           "T_ll": self.fixed_points["T_ll"],
                           "exergy_loss_rate": 0}


    def calculate_hex(self, fluid_numbers, h_dot_min_, h_out_w_, h_limit_s_=np.NAN,
                      points_=50, d_temp_separation_min_=0.5, calc_type_="const",
                      name_="evaporator", verbose_=False):
        """
        Utility function to calculate the T-h_dot curves and secondary fluid

        mass flow rates for a heat exchanger. it returns the state of the
        working fluid at the heat exchanger exit. while

        Parameters
        ----------
        fluid_numbers : list of two integers
            the working fluid index and the secondary fluid index.
        h_dot_min_ : float
            wanted heat flow rate [W].
        h_out_w_ : float
            specific enthalpy of the working fluid at the exit  [J/kg].
        h_limit_s_ : float, optional
            specific enthalpy of the secondary fluid at the exit  [J/kg].
            The default is np.NAN.
        points_ : integer, optional
            number of t-h points along the heat exchanger. The default is 50.
        d_temp_separation_min_ : float, optional
            the minimum approach temperature in K. The default is 0.5.
        calc_type_ : string, optional
            select the calculation type (only one implemented for now). The
            default is "const".
        name_ : string, optional
            name of the heat exchanger. The default is "evaporator".
        verbose_ : Boolean, optional
            additional printing/plotting. The default is False.

        Returns
        -------
        w_after_condenser : list
            actual output state of the working fluid [T,p,h,v,s,q].

        """
        w_actual, s_actual = [self.fluids[fluid_numbers[0]],
                              self.fluids[fluid_numbers[1]]]
        hex_act = \
            cb.hex_th.StaticHeatExchanger([w_actual,
                                           s_actual],
                                          h_dot_min_, h_out_w_, h_limit_s_,
                                          d_temp_separation_min=d_temp_separation_min_,
                                          name=name_)
        h_s_out = hex_act.find_pinch()
        m_dot_s, d_tempall, wf_states, sf_states =\
            hex_act.pinch_plot(plotting=verbose_)

        self.all_states.append(wf_states)
        self.all_states.append(sf_states)
        self.m_dots.append(hex_act.m_dot_wf)
        self.m_dots.append(m_dot_s)

        w_after_condenser = self.fluids[0].set_state([wf_states[0, 1],
                                                      wf_states[0, 2]], "PH")
        if verbose_:
            print("after condenser")
            w_actual.print_state()
        return w_after_condenser


    def calc_p_high(self, temps_wanted, verbose=False):
        """
        calculate a simple compression heat pump

        working with two sensible storages (or source and sink). At the moment
        the compressor model is just for a constant isentropic efficiency.
        An isenthalpic throttle is used for expnsion. The initial states are provided
        by the Fluid-instances. The pressures have to be selected carefully,
        they are not tested at the moment. The two desired storage temperatures
        are an input, and will be varied to meet the desired minmum approach
        temperatures in the heat exchangers.
        For optimization, the pressure levels of the cycle could be used within
        a second script. Several parameters are fixed
        along a calculation, this is in the dictionary "self.fixed_points".
        It can turn out that the foreseen evaporator exit temperature of the
        secondary(cooling) fluid is below the temperature of the isenthalpic
        temperature of the fluid coming out of the throttle. In this case the
        secondary exit temperature is increased.

        Parameters
        ----------
        temps_wanted : np.array([float,float])
            the two (high and low) temperatures to store at, may be varied
            towards the temperature of the surrounding, if the second law
            of thermodynamics requires this.
        verbose : Boolean, optional
            should values be printed? The deault is False

        Returns
        -------
        COP : flaot
            coefficient of performace.


        """

        state_in = copy.copy(self.fluids[0])
        w_actual = self.fluids[0]
        if verbose:
            print("Start:")
            w_actual.print_state()
        # compressor -----------------------
        w_after_compressor, work_c = \
            cb.compressor_simple.compressor(self.fixed_points["p_high"], self.fixed_points["eta_s"],
                                            self.fluids[0])
        if verbose:
            print("After compressor")
            w_actual.print_state()
       
        # condenser ------------------------                                       "PT")
        condenser = \
            cb.hex_th.StaticHeatExchanger([w_actual,
                                           self.fluids[1]],
                                          self.fixed_points["Q_dot_h"],
                                          self.fixed_points["h_h_out_w"],
                                          self.fixed_points["h_h_out_sec"],
                                          d_temp_separation_min=self.fixed_points["d_temp_min"],
                                          name="condenser")
        h_s_out = condenser.find_pinch()
        m_dot_s, d_tempall, wf_states, sf_states =\
            condenser.pinch_plot(plotting=False)
        m_dot_wf = self.fixed_points["Q_dot_h"] / \
            (wf_states[-1, 2] - wf_states[0, 2])
        self.all_states.append(wf_states)
        self.all_states.append(sf_states)
        self.m_dots.append(m_dot_wf)
        self.m_dots.append(m_dot_s)

        w_after_condenser = self.fluids[0].set_state([wf_states[0, 1],
                                                      wf_states[0, 2]], "PH")
        if verbose:
            print("after condenser")
            w_actual.print_state()

        power = work_c * m_dot_wf
        heat_flow_evap = self.fixed_points["Q_dot_h"] - power
        if verbose:
            print(
                f"power:{ power} W, heat-Evap. {heat_flow_evap} W\n")

        # throttle--------------------
        w_after_throttle = cb.throttle_simple.throttle(self.fixed_points["p_low"],
                                                       w_actual)

        if verbose:
            print("after throttle:")
            print(w_actual.print_state())
        w_act = w_actual.set_state(
            [w_after_throttle[1], w_after_throttle[2]], "PH")
        if self.fixed_points["T_ll"] < w_act[0] + self.fixed_points["d_temp_min"]:
            self.fixed_points["T_ll"] = w_act[0] + \
                self.fixed_points["d_temp_min"]
        if verbose:
            w_actual.print_state()

        evaporator = cb.hex_th.StaticHeatExchanger([w_actual, self.fluids[2]],
                                                   heat_flow_evap,
                                                   self.fixed_points["h_l_out_w"],
                                                   self.fixed_points["h_l_out_cold"],
                                                   d_temp_separation_min=self.fixed_points["d_temp_min"],
                                                   calc_type="const",
                                                   name="evaporator")
        factor_ev = evaporator.find_pinch()
        ms_ev, d_tempall_ec, w_ev, s_ev = evaporator.pinch_plot(
            plotting=False)

        self.all_states.append(w_ev)
        self.all_states.append(s_ev)
        self.m_dots.append(m_dot_wf)
        self.m_dots.append(ms_ev)


#####################################################
        w_after_evap = self.fluids[0].set_state([w_ev[0, 1], w_ev[0, 2]], "PH")
        q_evap = (w_after_evap[2] - w_after_throttle[2]) * m_dot_wf
        if heat_flow_evap - q_evap > 1:
            print(f"heat pump cycle not steady{heat_flow_evap, q_evap}")
        if verbose:
            print(f"  heat after Evaporator:{q_evap},power:{power}")
            w_actual.print_state()
            print("-test-", (sf_states[-1] - sf_states[0])*self.m_dots[1],
                  (s_ev[-1]-s_ev[0])*self.m_dots[-1])
        sec_fluids_states_in_out = np.array([[sf_states[0], sf_states[-1]],
                                            [s_ev[0], s_ev[-1]]])

        mass_flow_rates = np.array([self.m_dots[1], self.m_dots[-1]])

        ex_loss_rate = cb.exlo.exergy_loss_flow(sec_fluids_states_in_out,
                                                mass_flow_rates)
        ex_loss_alte = cb.exlo.exergy_loss_alternative(sec_fluids_states_in_out,
                                                       mass_flow_rates,
                                                       power_in=power)
        cop = self.fixed_points["Q_dot_h"] / power
        self.evaluation = {"Q_dot_h": self.fixed_points["Q_dot_h"],
                           "Power": power,
                           "p_high": self.fixed_points["p_high"],
                           "p_low": self.fixed_points["p_low"],
                           "exergy_loss_rate": ex_loss_rate,
                           "exergy_loss_alte": ex_loss_alte,
                           "sec_fluid_states_in_out": sec_fluids_states_in_out,
                           "sec_fluid_m_dots": mass_flow_rates,
                           "COP": cop
                           }
        return cop

    def hp_plot(self, f_name=_RESULTS_+"last_T_H_dot_plot_hp"):
        """
        plots the heat pump cycle and stores it to the given file (name)

        Returns
        -------
        None.

        """
        states = []
        connect_temp = []
        connect_h = []
        position = [0, -1]
        fig, ax = plt.subplots(1, 1)
        points = ["-k", "or", "-b", "vg"]
        m_dots = self.m_dots
        n_val = [2, 0]  # positions of h and T
        n_points = np.shape(self.all_states)[1]
        n_range = np.shape(m_dots)[0]  # how many streams? 2 x heat exchangers
        data = np.zeros((n_points, 4))
        names = ["HP-process", "{$H_{dot} /W", "T / K", "$m_{dot}"]
        all_data = []

        m_dots[3] = m_dots[3]  # / m_dots[2] * m_dots[0]
        m_dots[2] = m_dots[0]

        for i_process in range(4):

            cycle_states = self.all_states[i_process]
            cycle_states = cycle_states[cycle_states[:, 2].argsort(), :]
            if i_process == 2:
                liquid = self.fluids[0].set_state([cycle_states[0, 1], 0.],
                                                  "PQ")
                ax.plot((liquid[2] - cycle_states[0, 2]) * m_dots[i_process],
                        liquid[0], "o")
            cycle_states[:, 2] -= cycle_states[0, 2]

            cycle_states[:, 2] = cycle_states[:, 2] * m_dots[i_process]
            states.append(cycle_states)

            ax.plot(cycle_states[:, 2], cycle_states[:, 0], points[i_process])
            data[:, 0] = float(i_process)
            data[:, -1] = m_dots[i_process]
            data[:, 1:3] = cycle_states[:, n_val]
            all_data.append(copy.copy(data))
            if i_process == 0 or i_process == 2:
                connect_temp.append(cycle_states[position, 0])
                connect_h.append(cycle_states[position, 2])
        temperatures = np.array(connect_temp).reshape((4))
        enthalpies = np.array(connect_h).reshape((4))
        ax.set_xlabel("$\dot H$ / W")
        ax.set_ylabel("$T$ / K")

        for i_process in range(2):
            ax.plot([enthalpies[0+i_process], enthalpies[2+i_process]],
                    [temperatures[0+i_process], temperatures[2+i_process]], "r:")
        fig.savefig(f_name+".png")
        np.savetxt(f_name+".csv", np.concatenate(all_data),
                   delimiter=";", header=";".join(names))


if __name__ == "__main__":

    FLUID = "Propane * Butane * Pentane * Hexane"
    comp = [.75, 0.05, 0.15, 0.05]
    comp = [0.4,	0.3,	0.3, 0.0]  # [0.164,.3330,.50300,0.0]

    FLS = "Water"  #
    FLCOLD = "Methanol"  # "Water"  #

    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])

    coldFlm = cb.fprop.FluidModel(FLCOLD)
    coldFluid = cb.fprop.Fluid(coldFlm, [1.])

    # Condenser(c) and storage (s), secondary fluids fix all, temperatures(T in K),
    # pressures (p in Pa)
    _ETA_S_ = 0.67  # interesting when changed from 0.69 to 0.65, the efficiency
    # decreases, the reason is the low quality along throtteling then
    _STORAGE_T_IN_ = cb._T_SURROUNDING
    _COLD_STORAGE_T_IN_ = _STORAGE_T_IN_
    _STORAGE_T_OUT_ = 363.  # 395.0
    _COLD_STORAGE_T_OUT_ = 260.15
    _STORAGE_P_IN_ = 5e5
    _COLD_STORAGE_P_IN_ = 5e5
    _Q_DOT_MIN_ = 1e3  # and heat_flow rate (W)
    _D_T_SUPER_ = 5  # super heating of working fluid
    _D_T_MIN_ = 4.  # minimum approach temperature (pinch point)
    # high T-storages
    state_sec_out = secFluid.set_state([_STORAGE_T_OUT_, _STORAGE_P_IN_], "TP")
    state_sec_in = secFluid.set_state([_STORAGE_T_IN_, _STORAGE_P_IN_], "TP")

    #  low T sorages:
    state_cold_out = coldFluid.set_state(
        [_COLD_STORAGE_T_OUT_, _COLD_STORAGE_P_IN_], "TP")
    state_cold_in = coldFluid.set_state(
        [_COLD_STORAGE_T_IN_, _COLD_STORAGE_P_IN_], "TP")

    # working fluid
    T_DEW = _STORAGE_T_OUT_  # + _D_T_MIN_
    state_in_cond = myFluid.set_state([T_DEW, 1.], "TQ")  # find high pressure
    state_out_cond = myFluid.set_state([_STORAGE_T_IN_ + _D_T_MIN_,
                                        state_in_cond[1]], "TP")
    state_satv_evap = myFluid.set_state(
        [_STORAGE_T_IN_-_D_T_MIN_-_D_T_SUPER_, 1.], "TQ")  # find minimum pressure
    p_low = state_satv_evap[1]

    T_IN = _STORAGE_T_IN_ - _D_T_MIN_

    state_out_evap = myFluid.set_state([p_low,
                                        T_IN], "PT")

    FIXED_POINTS = {"eta_s": _ETA_S_,
                    "p_low": state_out_evap[1],
                    "p_high": state_in_cond[1],
                    "T_hh": _STORAGE_T_OUT_,
                    "h_h_out_sec": state_sec_out[2],
                    "h_h_out_w": state_out_cond[2],
                    "h_l_out_cold": state_cold_out[2],
                    "h_l_out_w": state_out_evap[2],
                    "T_hl": _STORAGE_T_IN_,
                    "T_lh": _STORAGE_T_IN_,
                    "T_ll": _COLD_STORAGE_T_OUT_,  # 256.0,
                    "Q_dot_h": _Q_DOT_MIN_,
                    "d_temp_min": _D_T_MIN_}

    print(
        f"p-ratio: {state_in_cond[1]/state_out_evap[1]: .2f}, p_low: {state_out_evap[1]/1e5: .2} bar")
    hp0 = HeatPump([myFluid, secFluid, coldFluid], FIXED_POINTS)
    cop = hp0.calc_p_high(FIXED_POINTS["p_high"], verbose=True)
    hp0.hp_plot()
    out = hp0.evaluation
    print(
        f"COP: {cop},p-ratio: {out['p_high']/out['p_low']:.2f}, p_low {out['p_low']/1e5:.2f} bar")
    print(
        f"COP: {cop},p-ratio: {out['p_high']/out['p_low']}, p_low {out['p_low']/1e5}")
    print(
        f'exergy loss rate: {out["exergy_loss_rate"]}, eff: {1-out["exergy_loss_rate"]/out["Power"]:.4f}')
    print(
        f'power= {out["Power"]:.2f} W, Alternative Ex-Loss:{out["exergy_loss_alte"]:.4f}')
    print(f'eff_alternative: {1-out["exergy_loss_alte"]/out["Power"]:.4f}')



# -*- coding: utf-8 -*-
"""
functions for compressor and expander output state calcultations

so far/here: only for fixed isentropic efficiencies

((Part of carbatpy.))
Created on Sun May 21 08:51:33 2023

@author: atakan
"""

import carbatpy as cb
import numpy as np


def compressor(p_out, eta_s, fluid, m_dot=1.0, calc_type="const_eta",
               name="compressor", plot_info={}):
    """
    compressor or expander output state calculation

    so far only for a constant isentropic efficiency, according to the pressure
    change an expansion or compression is detected and handled.

    Parameters
    ----------

    p_out : float
        output pressure.
    eta_s : float
        isentropic efficiency.
    fluid : fprop.Fluid
        entering fluid, including properties, composition, and model.
    m_dot : float, optional
        mass flow rate (in kg/s). Default is 1
    calc_type : string, optional
        how to calculate, so far, only one implemented. The default is
        "const_eta".
    name : string, optional
        name of the device. The default is "compressor".
    plot_info : dictionary, optional
        if not empty a Figure, an Axes, a list of What shall be plotted,
        a list with the colour/styles and a list with the labels must be
        passed. in "what", the two numbers coincide with the fluid THERMO
        order. The x-shift can be used in cycle calculations, to shift the
        curves, by the value (it will be added).
        The names in the dictionary are: "fig", "ax", "what","col",
        "label", "x-shift". Default is empty.

    Returns
    -------
    state_out : array of float
        compressor output state containing [T,p,h,v,s,q].
    work_specific : float
        work per kg of fluid, positive for compressor; units:J/kg.

    """
    state_in = fluid.properties.state
    expander = False
    if fluid.properties.pressure > p_out:
        expander = True

    if calc_type == "const_eta":
        fluid.set_state(
            [fluid.properties.entropy, p_out], "SP")

        diff_enthalpy_s = fluid.properties.enthalpy-state_in[2]

        if expander:
            work_specific = diff_enthalpy_s * eta_s
        else:
            work_specific = diff_enthalpy_s / eta_s

        state_out = fluid.set_state([state_in[2] + work_specific, p_out], "HP")

    else:
        raise Exception(
            f"The option{calc_type} is not yet implemented for compressors")

    power = m_dot * work_specific
    plot_temp_h_flow(state_in, state_out, m_dot, plot_info)
    #

    return state_out, work_specific, power


def pump(p_out, eta_s, fluid, m_dot=1.0, calc_type="const_eta", name="pump",
         plot_info={}):
    """
    Calculate the exit state of a pump assuming an incompressible fluid.

    Only formulated for constant isentropic efficiency

    Parameters
    ----------
    p_out : float
        output pressure.
    eta_s : float
        isentropic efficiency.
    fluid : fprop.Fluid
        entering fluid, including properties, composition, and model.
    m_dot : float, optional
        mass flow rate (in kg/s). Default is 1
    calc_type : string, optional
        how to calculate, so far, only one implemented. The default is
        "const_eta".
    name : string, optional
        name of the device. The default is "pump".
    plot_info : dictionary, optional
        if not empty a Figure, an Axes, a list of What shall be plotted,
        a list with the colour/styles and a list with the labels must be
        passed. in "what", the two numbers coincide with the fluid THERMO
        order. The x-shift can be used in cycle calculations, to shift the
        curves, by the value (it will be added).
        The names in the dictionary are: "fig", "ax", "what","col",
        "label", "x-shift". Default is empty.

    Returns
    -------
    state_out : array of float
        compressor output state containing [T,p,h,v,s,q].
    work_specific : float
        work per kg of fluid, positive for compressor; units:J/kg.

    """
    state_in = fluid.properties.state
    if calc_type == "const_eta":

        work_is = state_in[3] * (p_out - state_in[1])
        if work > 0:
            work_specific = work_is / eta_s
        else:
            work_specific = work_is * eta_s
        h_out = state_in[2] + work_specific
        state_out = fluid.set_state([h_out, p_out], "HP")
    else:
        raise Exception(
            f"The option{calc_type} is not yet implemented for pumps")
    power = m_dot * work_specific

    plot_temp_h_flow(state_in, state_out, m_dot, plot_info)


    return state_out, work_specific, power

def plot_temp_h_flow(_state_in, _state_out, _m_dot, _plot_info):
    """
    plotting a T-H-dot diagram for simple flows (compressor, throttle etc.)

    Parameters
    ----------
    _state_in : np.array
        entering state [T,p,h,v,s,...].
    _state_out : np.array
        exiting state.
    _m_dot : float
        mass flow rate (kg/s).
    _plot_info : dictionary
        if not empty a Figure, an Axes, a list of What shall be plotted,
        a list with the colour/styles and a list with the labels must be
        passed. in "what", the two numbers coincide with the fluid THERMO
        order. The x-shift can be used in cycle calculations, to shift the
        curves, by the value (it will be added).
        The names in the dictionary are: "fig", "ax", "what","col",
        "label", "x-shift".

    Returns
    -------
    None.

    """
    if len(_plot_info) > 0:
        if _plot_info["what"][0] ==2:
            data = np.array([_state_in[_plot_info["what"][0]] ,
                                  _state_out[_plot_info["what"][0]]]) * _m_dot \
                + _plot_info["x-shift"][0]
            _plot_info["ax"].plot(data,
                                 [_state_in[_plot_info["what"][1]],
                                  _state_out[_plot_info["what"][1]]],
                                 _plot_info["col"][0],
                                 label=_plot_info["label"][0])
        else:
            print(f"Pump: plotting only implemented fot T-H_dot [2,0]. You requested{_plot_info['what']}")




if __name__ == "__main__":
    import matplotlib.pyplot as plt

    FLUID = "Propane * Pentane"
    comp = [.80, 0.2]
    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)
    P_LOW = 1e5
    T_IN = 310.
    DT_IN_LIQ = -5
    state_in_act = myFluid.set_state([T_IN, P_LOW], "TP")
    P_OUT = 10e5
    ETA_S = .7
    M_DOT = 1e-3
    fig0, ax0 = plt.subplots()
    PLOT_INFO ={"fig":fig0, "ax":ax0, "what":[2,0],"col":["r:","k"],
    "label":["compressor","xx"], "x-shift":[0,0]}

    # Compressor-------------
    state_o, work, power_c = compressor(P_OUT, ETA_S, myFluid, m_dot=M_DOT,
                                      plot_info=PLOT_INFO)
    print(myFluid.properties.temperature, work)
    print("\nCompressor", state_in_act, "\n", state_o, "\n", state_o-state_in_act)
    PLOT_INFO["col"]=["k",""]
    PLOT_INFO["label"]=["expander",""]
    state_in_c = state_o
    state_o, work, power_e = compressor(P_LOW, ETA_S, myFluid, m_dot=M_DOT,
                                      plot_info=PLOT_INFO)
    print("\nExpander:", state_in_c, "\n", state_o, "\n", state_o-state_in_c, work)

    # Pump, incompressible:

    state_in_p = myFluid.set_state([P_LOW, 0], "PQ")
    state_in_p = myFluid.set_state([P_LOW, state_in_p[0]+DT_IN_LIQ], "PT")
    PLOT_INFO["col"]=["bv:",""]
    PLOT_INFO["label"]=["pump",""]
    state_o, work, power = pump(P_OUT, ETA_S, myFluid, m_dot=M_DOT,
                                plot_info=PLOT_INFO)
    print(f"pump work: {work:.3f} J/kg, state:{state_o}")
    ax0.legend()
    ax0.set_ylabel("T/K")
    ax0.set_xlabel("$\dot h$/ kW")



# -*- coding: utf-8 -*-
"""
Some constants for carbatpy usage are set here

* The two (!) Refprop installation directories (with dll, fluids etc.)
* he directory, where the results shall be written.

Created on Sun Nov  5 16:01:02 2023

@author: atakan
"""

import os
# global _T_SURROUNDING, _P_SURROUNDING, _RESULTS_DIR, _CARBATPY_BASE_DIR
_T_SURROUNDING = 288.15 # K
_P_SURROUNDING =1.013e5  # Pa
_TREND = {"USE_TREND":False, "TREND_DLL":"", 'TREND_PATH':""}

directory = os.getcwd()
try:
    _CARBATPY_BASE_DIR = os.environ["CARBATPY_BASE_DIR"]
except:
    _CARBATPY_BASE_DIR = directory

# The two installations of REFPROP , one for the working fluid
# and one for the secondary fluid. With only one installation, the instances
# mix up ...
os.environ['RPPREFIX'] = r'C:/Program Files (x86)/REFPROP'
os.environ['RPPREFIXs'] = r'C:/Program Files (x86)/REFPROP/secondCopyREFPROP'
if _TREND["USE_TREND"]:
    try:
        _TREND["TREND_DLL"] = os.environ['TREND_DLL']
        _TREND["TREND_PATH"] = os.environ['TREND_PATH']
    except:
        print("Trend not found! Check the environmentvariable TREND_DLL, TREND_PATH")
        _TREND_DLL = ""
        _TREND_PATH = ''
    
try:
    _RESULTS_DIR = os.environ['CARBATPY_RES_DIR']
except:
    try:
        _RESULTS_DIR = os.environ['TEMP']
    except:
        try:
            
            _RESULTS_DIR = directory + r"\\tests\\test_files"
        except Exception as no_file:
            print("Please set the envirionment variable: CARBATPY_RES_DIR !", no_file)


# -*- coding: utf-8 -*-
"""
Created on Fri Dec 15 14:33:04 2023

@author: atakan
"""

import datetime
import pandas as pd
import graphviz as gv

DIRNAME = "..//..//data"
FILE_NAME = "orc_structure.xlsx"


def plot_cycle_structure(fname, dirname=DIRNAME, cycle_name="cycle",
                         format_="png",
                         engine="circo"):
    """
    read an Excel file with information about the cycle structure and use graphviz

    for visualization, a file is generated.

    Parameters
    ----------
    fname : string
        the excel filename.
    dirname : string, optional
        the directory name to read the file and to store afterwards. The default is DIRNAME.
    cycle_name : string, optional
        name of the cycle. The default is "cyle".
    format : String, optional
        plot-format (png, pdf etc). The default is "png".
    engine : string, optional
        an engine name from graphviz (dot, circo etc.). The default is "circo".

    Returns
    -------
    None.

    """

    verbose = False
    fname_exel = dirname + "//" + fname
    fname_out = dirname + "//" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") +\
        cycle_name

    data = pd.read_excel(fname_exel, index_col=0)
    dot = gv.Digraph(cycle_name, comment='ORC in a Carnot Battery')
    dot.engine = engine
    dot.attr(rankdir="TB")

    for line_n, node_name in enumerate(data["Name(unique)"]):
        if data["in_cycle"].iloc[line_n]:
            shape = "box"
            style = "filled"
            color = "yellow"
            arrowhead = "normal"
            acolor = "red"
        else:
            shape = "egg"
            style = "filled"
            color = "cyan"
            arrowhead = "vee"
            acolor = "blue"

        dot.node(node_name, shape=shape, style=style, color=color)
        if pd.notnull(data["input1"].iloc[line_n]):
            label = str(line_n)
            dot.edge(data["input1"].iloc[line_n], node_name, label=label,
                     arrowhead=arrowhead,
                     color=acolor)
        if pd.notnull(data["output1"].iloc[line_n]):
            if not data["in_cycle"].iloc[line_n]:
                dot.edge(node_name, data["output1"].iloc[line_n],
                         label=str(line_n),
                         arrowhead=arrowhead,
                         color=acolor)

    if verbose:
        print(dot.source, fname_out)
    dot.render(format=format_, filename=fname_out, view=True)


if __name__ == "__main__":
    plot_cycle_structure(FILE_NAME, cycle_name="orc")
    plot_cycle_structure("hp_structure.xlsx", cycle_name="heat_pump")



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 13 09:47:06 2023

@author: atakan
"""

import carbatpy as cb

_RESULTS_ = r"C:\Users\atakan\sciebo\results\\"


FLUID = "Propane * Ethane * Pentane *Isobutane"
#comp = [.75, 0.05, 0.15, 0.05]
x1 =0.25
x2 =0.0
x3 =0.0
x4 = 1- x1-x2-x3
comp = [x1,x2, x3, x4]  # [0.164,.3330,.50300,0.0]
print(f"{FLUID}, composition:\n{comp}")
T_SUR =273.15+20
FLS = "Water"  #
FLCOLD = "Water"  #"Methanol"  # "Water"  #

flm = cb.fprop.FluidModel(FLUID)
myFluid = cb.fprop.Fluid(flm, comp)

secFlm = cb.fprop.FluidModel(FLS)
secFluid = cb.fprop.Fluid(secFlm, [1.])

coldFlm = cb.fprop.FluidModel(FLCOLD)
coldFluid = cb.fprop.Fluid(coldFlm, [1.])

# Condenser(c) and storage (s), secondary fluids fix all, temperatures(T in K),
# pressures (p in Pa)
_ETA_S_ = 0.57  # interesting when changed from 0.69 to 0.65, the efficiency
# decreases, the reason is the low quality along throtteling then
_STORAGE_T_IN_ = 273.15 + 50
_COLD_STORAGE_T_IN_ = T_SUR
_STORAGE_T_OUT_ = 273.15 + 60  # 395.0
_COLD_STORAGE_T_OUT_ = 284.15
_STORAGE_P_IN_ = 5e5
_COLD_STORAGE_P_IN_ = 5e5
_Q_DOT_MIN_ = 1e3  # and heat_flow rate (W)
_D_T_SUPER_ = 15  # super heating of working fluid
_D_T_MIN_ = 4.  # minimum approach temperature (pinch point)
# high T-storages
state_sec_out = secFluid.set_state([_STORAGE_T_OUT_, _STORAGE_P_IN_], "TP")
state_sec_in = secFluid.set_state([_STORAGE_T_IN_, _STORAGE_P_IN_], "TP")

#  low T sorages:
state_cold_out = coldFluid.set_state(
    [_COLD_STORAGE_T_OUT_, _COLD_STORAGE_P_IN_], "TP")
state_cold_in = coldFluid.set_state(
    [_COLD_STORAGE_T_IN_, _COLD_STORAGE_P_IN_], "TP")

# working fluid
T_DEW = _STORAGE_T_OUT_ #+ _D_T_MIN_
state_in_cond = myFluid.set_state([T_DEW, 1.], "TQ")  # find high pressure
state_out_cond = myFluid.set_state([_STORAGE_T_IN_ + _D_T_MIN_,
                                    state_in_cond[1]], "TP")
state_satv_evap = myFluid.set_state(
    [_STORAGE_T_IN_-_D_T_MIN_-_D_T_SUPER_, 1.], "TQ")  # find minimum pressure
p_low = state_satv_evap[1]

T_IN = _STORAGE_T_IN_ - _D_T_MIN_

state_out_evap = myFluid.set_state([p_low,
                                    T_IN], "PT")

FIXED_POINTS = {"eta_s": _ETA_S_,
                "p_low": state_out_evap[1],
                "p_high": state_in_cond[1],
                "T_hh": _STORAGE_T_OUT_,
                "h_h_out_sec": state_sec_out[2],
                "h_h_out_w": state_out_cond[2],
                "h_l_out_cold": state_cold_out[2],
                "h_l_out_w": state_out_evap[2],
                "T_hl": _STORAGE_T_IN_,
                "T_lh": _STORAGE_T_IN_,
                "T_ll": _COLD_STORAGE_T_OUT_,  # 256.0,
                "Q_dot_h": _Q_DOT_MIN_,
                "d_temp_min": _D_T_MIN_}

print(
    f"p-ratio: {state_in_cond[1]/state_out_evap[1]: .2f}, p_low: {state_out_evap[1]/1e5: .2} bar")
hp0 = cb.hp_simple.HeatPump([myFluid, secFluid, coldFluid], FIXED_POINTS)
cop = hp0.calc_p_high(FIXED_POINTS["p_high"], verbose=False)
hp0.hp_plot()
out = hp0.evaluation
print(
    f"COP: {cop},p-ratio: {out['p_high']/out['p_low']:.2f}, p_low {out['p_low']/1e5:.2f} bar")
print(
    f"COP: {cop},p-ratio: {out['p_high']/out['p_low']}, p_low {out['p_low']/1e5}")
print(
    f'exergy loss rate: {out["exergy_loss_rate"]}, eff: {1-out["exergy_loss_rate"]/out["Power"]:.4f}')
